/******************************************************************************

Copyright 2003, 2004 Christian Gollan

This file is part of the W2D package.

The W2D package is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

The W2D package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with the W2D package; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <string.h>
#include "IMG.h"
#include "W2D.h"
#include "config.h"
#ifdef HAVE_LIBZ
#include <zlib.h>
#endif

int
IMG_get_depth(IMG* img)
{
  int img_depth;

  for(img_depth=0;img->dis_mode[img_depth]!=-2;img_depth++);

  return(img_depth-1);
}

int
IMG_get_depth_list(IMG_list* img_list)
{
  int img_depth=-1;
  int img_nr;

  for(img_nr=0;img_nr<img_list->total_nr;img_nr++)
    if( img_depth < IMG_get_depth(&(img_list->img[img_nr])) )
      img_depth=IMG_get_depth(&(img_list->img[img_nr]));

  return(img_depth);
}

int
IMG_get_max_img_dim_list(IMG_list* img_list)
{
  int img_nr;
  int max_img_dim=0;
  
  for(img_nr=0;img_nr<img_list->total_nr;img_nr++)
    if( img_list->img[img_nr].x_size*img_list->img[img_nr].y_size > 
	max_img_dim)
      max_img_dim=img_list->img[img_nr].x_size*img_list->img[img_nr].y_size;
  
  return(max_img_dim);
}

int
IMG_get_max_img_x_size_list(IMG_list* img_list)
{
  int img_nr;
  int max_x_size=0;
  
  for(img_nr=0;img_nr<img_list->total_nr;img_nr++)
    if(img_list->img[img_nr].x_size > max_x_size)
	max_x_size=img_list->img[img_nr].x_size;

  return(max_x_size);
}

int
IMG_get_max_img_y_size_list(IMG_list* img_list)
{
  int img_nr;
  int max_y_size=0;
  
  for(img_nr=0;img_nr<img_list->total_nr;img_nr++)
    if(img_list->img[img_nr].y_size > max_y_size)
	max_y_size=img_list->img[img_nr].y_size;

  return(max_y_size);
}

int
IMG_get_min_img_dim_list(IMG_list* img_list)
{
  int img_nr;
  int min_img_dim=IMG_get_max_img_dim_list(img_list);
  
  for(img_nr=0;img_nr<img_list->total_nr;img_nr++)
    if( img_list->img[img_nr].x_size*img_list->img[img_nr].y_size < 
	min_img_dim)
      min_img_dim=img_list->img[img_nr].x_size*img_list->img[img_nr].y_size;

  return(min_img_dim);
}

int
IMG_get_min_img_x_size_list(IMG_list* img_list)
{
  int img_nr;
  int min_x_size=IMG_get_max_img_x_size_list(img_list);
  
  for(img_nr=0;img_nr<img_list->total_nr;img_nr++)
    if(img_list->img[img_nr].x_size < min_x_size)
	min_x_size=img_list->img[img_nr].x_size;

  return(min_x_size);
}

int
IMG_get_min_img_y_size_list(IMG_list* img_list)
{
  int img_nr;
  int min_y_size=IMG_get_max_img_y_size_list(img_list);
  
  for(img_nr=0;img_nr<img_list->total_nr;img_nr++)
    if(img_list->img[img_nr].y_size < min_y_size)
	min_y_size=img_list->img[img_nr].y_size;

  return(min_y_size);
}

//Read from a *.jf-file images which will be stored where feature
//points at and the class of each image will be stored from the beginning
//of pointer klasse. dim will be set to the dimension of the images and
//the return value describes how many images have been readn 
int
IMG_read_list (char *filename, int *dim, int anz_bilder, int **klasse,
	       double ***feature)
{
  int i, j;
  int anz_klassen;
  FILE *fp;
  float temp;

  if ((fp = fopen (filename, "r")) == NULL)
    {
      fprintf (stderr, "can't open file %s\n", filename);
      return (0);
    }
  else
    {
      //*******calculate how many images are in this file************
      fscanf (fp, "%d %d\n", &anz_klassen, dim);

      anz_klassen = 0;
      while (fscanf (fp, "%f", &temp) != EOF)
	anz_klassen++;
      anz_klassen = anz_klassen / (*dim + 1);
      if (anz_bilder == 0)
	anz_bilder = anz_klassen;
      if (anz_bilder > anz_klassen)
	{
	  fprintf (stderr, "file %s has wrong image numbers: %d != %d\n",
		   filename, anz_klassen, anz_bilder);
	  fclose (fp);
	  return (0);
	}
      fseek (fp, 0, SEEK_SET);
      //***********************************************************

      fscanf (fp, "%d %d\n", &anz_klassen, dim);
      *klasse = (int *) malloc (anz_bilder * sizeof (int));
      if (klasse == NULL)
	{
	  fprintf (stderr, "not enough memory for images from File %s\n",
		   filename);
	  fclose (fp);
	  return (0);
	}

      *feature = malloc (anz_bilder * sizeof (double *));
      if (*feature == NULL)
	{
	  fprintf (stderr, "not enough memory for images from File %s\n",
		   filename);
	  free (klasse);
	  fclose (fp);
	  return (0);
	}

      for (i = 0; i < anz_bilder; i++)
	{
	  fscanf (fp, "%d", &((*klasse)[i]));
	  (*feature)[i] = malloc (*dim * sizeof (double));
	  for (j = 0; j < *dim; j++)
	    {
	      fscanf (fp, "%f", &temp);
	      (*feature)[i][j] = (double) temp;
	    }
	  fscanf (fp, "\n");
	}
      fprintf (stdout,
	       "$IMG_read_list(%s,...);\n#read %d images, dimension %d\n",
	       filename, anz_bilder, *dim);
    }
  fclose (fp);
  return (anz_bilder);
}

IMG_list*
IMG_load_list_ejf (char* filename)
{
#ifdef HAVE_LIBZ
  gzFile fp;
#else
  FILE* fp;
#endif
  int img_depth;
  int img_nr;
  IMG_list *img_list;
  int char_byte;
  unsigned char* buffer;
  int buffer_length;
  int buffer_pos;
  int line_nr;
  int column_pos;
  int class=0;
  int int_temp;
  double double_temp;

#ifdef HAVE_LIBZ
  if((fp = gzopen (filename, "r")) == NULL)
#else
  if((fp = fopen (filename, "r")) == NULL)
#endif
    {
      fprintf (stderr, "can't open file %s\n", filename);
      exit (-1);
    }    

  img_list = malloc (sizeof (IMG_list));
  if (img_list == NULL ||
      (img_list->name=(char*) malloc((strlen(filename)+1)*sizeof(char))) == NULL)			   
    {
      fprintf (stderr, "not enough memory for images from file %s\n",
	       filename);
      exit(-1);
    }
  strcpy(img_list->name, filename);
  img_list->total_nr=0;
  img_list->img=NULL;
 
  buffer=NULL;
  buffer_length=1;
  buffer_pos=0;
  char_byte=0;
  line_nr=1;
  column_pos=1;
  img_depth=-1;
  img_nr=0;
  while(char_byte!=-1)
    {
#ifdef HAVE_LIBZ
  char_byte = gzgetc(fp);
#else
  char_byte = fgetc(fp);
#endif
    if(char_byte==' ' || char_byte=='\t' 
	 || char_byte=='\n' || char_byte==-1)
	{
	  if(buffer_pos>0)
	    {
	      buffer[buffer_pos]='\0';
	      if(column_pos<=3)
		{
		  if(sscanf(buffer, "%d%*s", &int_temp) != 1)
		    {
		      fprintf(stderr,"expect integer in line:%d column:%d of file %s\n",
			      line_nr,column_pos,filename);
		      exit(-1);
		    }
		}
	      else
		{
		  if(sscanf(buffer, "%lf%*s", &double_temp) != 1)
		    {
		      fprintf(stderr,"expect double in line:%d column:%d of file %s\n",
			      line_nr,column_pos,filename);
		      exit(-1);
		    }		
		}
	      buffer_pos=0;

	      switch(column_pos)
		{
		case 1:
		  class=int_temp;
		  break;
		case 2:
		  if(img_depth==-1)
		    {
		      fprintf(stderr,"expect only one integer for first line file %s\n",
			      filename);
		      exit(-1);
		    }		    
		  if(img_depth==0)
		    {
		      img_list->total_nr++;
		      img_list->img = realloc (img_list->img,img_list->total_nr*sizeof (IMG));
		      if (img_list->img == NULL)
			{
			  fprintf (stderr, "not enough memory for images from file %s\n",
				   filename);
			  exit (-1);
			}			  
		      img_list->img[img_nr].data_list = malloc (sizeof (double *) * 2);
		      img_list->img[img_nr].data_length = malloc (sizeof (int) * 2);
		      img_list->img[img_nr].dis_mode = malloc (sizeof (int) * 2);
		      img_list->img[img_nr].dis_scale = malloc (sizeof (double) * 2);
		      if (img_list->img[img_nr].data_list == NULL
			  || img_list->img[img_nr].data_length == NULL
			  || img_list->img[img_nr].dis_mode == NULL
			  || img_list->img[img_nr].dis_scale == NULL)
			{
			  fprintf (stderr, "not enough memory for images from file %s\n",
				   filename);
			  exit(-1);
			}  
		      img_list->img[img_nr].class=class;
		      img_list->img[img_nr].x_size=int_temp;
		    }
		  img_list->img[img_nr].data_length[img_depth]=int_temp;    
		  break;
		case 3:
		  if(img_depth==0)
		    img_list->img[img_nr].y_size=int_temp;
		  img_list->img[img_nr].data_length[img_depth]*=int_temp;
		  img_list->img[img_nr].dis_mode[img_depth] = -1;
		  img_list->img[img_nr].data_list[img_depth] =
		    malloc(img_list->img[img_nr].data_length[img_depth]
			   *sizeof(double));
		  if(img_list->img[img_nr].data_list[img_depth]==NULL)
		    {
		      fprintf (stderr, "not enough memory for images from file %s\n",
			       filename);
		      exit(-1);
		    }
		  break;
		default:
		  if((column_pos-4)>=img_list->img[img_nr].data_length[img_depth])
		    {
		      fprintf(stderr,"expect newline in line:%d column:%d of file %s\n",
			      line_nr,column_pos,filename);
		      exit(-1);
		    }
		  img_list->img[img_nr].data_list[img_depth][column_pos-4]=double_temp;
		}
	      column_pos++;
	    }
	  if(char_byte=='\n')
	    {
	      if((column_pos>=4 || column_pos==1 || column_pos==3) && class!=-1)
		{
		  if((column_pos-4)!=img_list->img[img_nr].data_length[img_depth]
		     || column_pos==1 || column_pos==3)
		    {
		      fprintf(stderr,"unexpected length of line:%d of file %s\n",
			      line_nr,filename);
		      exit(-1);
		    }
		  img_nr++;
		}
	      if(column_pos==2 && class!=-1)
		img_depth++;
	      if(column_pos==2 && img_depth>0)
		{
		  for(img_nr=0;img_nr<img_list->total_nr;img_nr++)
		    {
		      img_list->img[img_nr].data_list = 
			realloc (img_list->img[img_nr].data_list, sizeof (double *)*(img_depth+2));
		      img_list->img[img_nr].data_length = 
			realloc (img_list->img[img_nr].data_length, sizeof (int)*(img_depth+2));
		      img_list->img[img_nr].dis_mode = 
			realloc (img_list->img[img_nr].dis_mode, sizeof (int)*(img_depth+2));
		      img_list->img[img_nr].dis_scale = 
			realloc (img_list->img[img_nr].dis_scale, sizeof (double)*(img_depth+2));
		      if (img_list->img[img_nr].data_list == NULL
			  || img_list->img[img_nr].data_length == NULL
			  || img_list->img[img_nr].dis_mode == NULL
			  || img_list->img[img_nr].dis_scale == NULL)
			{
			  fprintf (stderr, "not enough memory for images from file %s\n",
				   filename);
			  exit(-1);
			}  			  
		    }
		  img_nr=0;
		}
	      line_nr++;
	      column_pos=1;
	    }
	}
      else
	{
	  if(buffer_length<buffer_pos+2)
	    {
	      buffer_length*=2;
	      buffer=(unsigned char*) realloc(buffer,buffer_length * sizeof(unsigned char));
		if (buffer == NULL)
		  {
		    fprintf (stderr, "not enough memory for images from File %s\n",
			     filename);
		    exit(-1);
		  }
	    }
	  buffer[buffer_pos]=(unsigned char) char_byte;
	  buffer_pos++;
	}
    }

  for(img_nr=0;img_nr<img_list->total_nr;img_nr++)
    { 
      int_temp=snprintf(NULL,0,"%d",img_nr)+1;
      img_list->img[img_nr].name=malloc(sizeof(char) * int_temp);
      if (img_list->img[img_nr].name == NULL)
	{
	  fprintf (stderr, "not enough memory for images from file %s\n",
		   filename);
	  return (NULL);
	}
      snprintf(img_list->img[img_nr].name,int_temp,"%d",img_nr);
      img_list->img[img_nr].data_list[img_depth+1] = NULL;
      img_list->img[img_nr].dis_mode[img_depth+1] = -2;
    }
#ifdef HAVE_LIBZ
  gzclose(fp);
#else
  fclose(fp);
#endif

  fprintf (stdout,
	   "$IMG_load_list_ejf(%s);\n#read %d images, img_depth: %d\n",
	   filename, img_list->total_nr, img_depth); 

  free(buffer);
  return(img_list); 
}

IMG_list*
IMG_load_list_jf (char* filename)
{
#ifdef HAVE_LIBZ
  gzFile fp;
#else
  FILE* fp;
#endif
  int img_nr;
  IMG_list *img_list;
  int char_byte;
  unsigned char* buffer;
  int buffer_length;
  int buffer_pos;
  int line_nr;
  int column_pos;
  int class=0;
  int int_temp;
  double double_temp;
  int dimension;

#ifdef HAVE_LIBZ
  if((fp = gzopen (filename, "r")) == NULL)
#else
  if((fp = fopen (filename, "r")) == NULL)
#endif
    {
      fprintf (stderr, "can't open file %s\n", filename);
      exit (-1);
    }    

  img_list = malloc (sizeof (IMG_list));
  if (img_list == NULL ||
      (img_list->name=(char*) malloc((strlen(filename)+1)*sizeof(char))) == NULL)			   
    {
      fprintf (stderr, "not enough memory for images from file %s\n",
	       filename);
      exit(-1);
    }
  strcpy(img_list->name, filename);
  img_list->total_nr=0;
  img_list->img=NULL;
 
  buffer=NULL;
  buffer_length=1;
  buffer_pos=0;
  char_byte=0;
  line_nr=1;
  column_pos=1;
  img_nr=0;
  dimension=-1;
  while(char_byte!=-1)
    {
#ifdef HAVE_LIBZ
    char_byte = gzgetc(fp);
#else
    char_byte = fgetc(fp);
#endif
    if(char_byte==' ' || char_byte=='\t' 
	 || char_byte=='\n' || char_byte==-1)
	{
	  if(buffer_pos>0)
	    {
	      buffer[buffer_pos]='\0';
	      if(line_nr==1 && column_pos==2)
		{
		  if(sscanf(buffer, "%d%*s", &dimension) != 1)
		    {
		      fprintf(stderr,"expect integer in line:%d column:%d of file %s\n",
			      line_nr,column_pos,filename);
		      exit(-1);
		    }
		}
	      else
		if(column_pos<=1)
		  {
		    if(sscanf(buffer, "%d%*s", &int_temp) != 1)
		      {
			fprintf(stderr,"expect integer in line:%d column:%d of file %s\n",
				line_nr,column_pos,filename);
			exit(-1);
		      }
		  }
		else
		  {
		    if(sscanf(buffer, "%lf%*s", &double_temp) != 1)
		      {
			fprintf(stderr,"expect double in line:%d column:%d of file %s\n",
				line_nr,column_pos,filename);
			exit(-1);
		      }		
		  }
	      buffer_pos=0;
	      if(line_nr>1 && dimension==-1)
		{
		  fprintf(stderr,"first line second column didn't have dimension in file %s\n",
			  filename);
		  exit(-1);		  
		}
	      if(line_nr>1)
		switch(column_pos)
		  {
		  case 1:
		    class=int_temp;		    
		    if(class!=-1)
		      {
			img_list->total_nr++;
			img_list->img = realloc (img_list->img,img_list->total_nr*sizeof (IMG));
			if (img_list->img == NULL)
			  {
			    fprintf (stderr, "not enough memory for images from file %s\n",
				     filename);
			    exit (-1);
			  }			  
			img_list->img[img_nr].data_list = malloc (sizeof (double *) * 2);
			img_list->img[img_nr].data_length = malloc (sizeof (int) * 2);
			img_list->img[img_nr].dis_mode = malloc (sizeof (int) * 2);
			img_list->img[img_nr].dis_scale = malloc (sizeof (double) * 2);
			if (img_list->img[img_nr].data_list == NULL
			    || img_list->img[img_nr].data_length == NULL
			    || img_list->img[img_nr].dis_mode == NULL
			    || img_list->img[img_nr].dis_scale == NULL)
			  {
			    fprintf (stderr, "not enough memory for images from file %s\n",
				     filename);
			    exit(-1);
			  }  
			img_list->img[img_nr].class=class;
			img_list->img[img_nr].x_size=(int)sqrt(dimension);
			img_list->img[img_nr].y_size=(int)sqrt(dimension);		      
			img_list->img[img_nr].data_length[0]=dimension;    
			img_list->img[img_nr].dis_mode[0] = -1;
			img_list->img[img_nr].data_list[0] =
			  malloc(dimension*sizeof(double));
			if(img_list->img[img_nr].data_list[0]==NULL)
			  {
			    fprintf (stderr, "not enough memory for images from file %s\n",
				     filename);
			    exit(-1);
			  }
		      }
		    break;
		  default:
		    if((column_pos-2)>=img_list->img[img_nr].data_length[0])
		      {
			fprintf(stderr,"expect newline in line:%d column:%d of file %s\n",
				line_nr,column_pos,filename);
			exit(-1);
		      }
		    img_list->img[img_nr].data_list[0][column_pos-2]=double_temp;
		  }
	      column_pos++;
	      
	    }
	  if(char_byte=='\n')
	    {
	      if(column_pos>=2 && line_nr>1 && class!=-1)
		{
		  if((column_pos-2)!=img_list->img[img_nr].data_length[0])
		    {
		      fprintf(stderr,"unexpected length of line:%d of file %s\n",
			      line_nr,filename);
		      exit(-1);
		    }
		  img_nr++;
		}
	      line_nr++;
	      column_pos=1;
	    }
	    
	}
    else
      {
	if(buffer_length<buffer_pos+2)
	  {
	    buffer_length*=2;
	    buffer=(unsigned char*) realloc(buffer,buffer_length * sizeof(unsigned char));
	    if (buffer == NULL)
	      {
		fprintf (stderr, "not enough memory for images from File %s\n",
			 filename);
		exit(-1);
	      }
	  }
	buffer[buffer_pos]=(unsigned char) char_byte;
	buffer_pos++;
      }
    }
  
  for(img_nr=0;img_nr<img_list->total_nr;img_nr++)
    { 
      int_temp=snprintf(NULL,0,"%d",img_nr)+1;
      img_list->img[img_nr].name=malloc(sizeof(char) * int_temp);
      if (img_list->img[img_nr].name == NULL)
	{
	  fprintf (stderr, "not enough memory for images from file %s\n",
		   filename);
	  return (NULL);
	}
      snprintf(img_list->img[img_nr].name,int_temp,"%d",img_nr);
      img_list->img[img_nr].data_list[1] = NULL;
      img_list->img[img_nr].dis_mode[1] = -2;
    }
#ifdef HAVE_LIBZ
  gzclose(fp);
#else
  fclose(fp);
#endif

  fprintf (stdout,
	   "$IMG_load_list_jf(%s);\n#read %d images\n",
	   filename, img_list->total_nr); 
  free(buffer);
  return(img_list); 
}

IMG_list *
IMG_load_list(char *filename)
{
  IMG_list *img_list;
  
  if(strlen(filename)>=4 && strcmp(".ejf",&(filename[strlen(filename)-4]))==0)
    img_list=IMG_load_list_ejf(filename);
  else
    if(strlen(filename)>=3 && strcmp(".jf",&(filename[strlen(filename)-3]))==0) 
      img_list=IMG_load_list_jf(filename);  
    else
      if(strlen(filename)>=6 && strcmp(".jf.gz",&(filename[strlen(filename)-6]))==0)
	{
#ifndef HAVE_LIBZ
	  fprintf(stderr,"not build with zlib for .jf.gz support\n");
	  exit(-1);
#endif
	  img_list=IMG_load_list_jf(filename);
	}
      else
	if(strlen(filename)>=7 && strcmp(".ejf.gz",&(filename[strlen(filename)-7]))==0) 
	  {
#ifndef HAVE_LIBZ
	    fprintf(stderr,"not build with zlib for .ejf.gz support\n");
	    exit(-1);
#endif
	    img_list=IMG_load_list_ejf(filename);
	  }
	else
	  {
	    fprintf(stderr,"filename %s has to end with .jf .ejf .jf.gz or .ejf.gz\n",filename);
	    exit(-1);
	  }
  if(img_list==NULL)
    exit(-1);
  
  return(img_list);
}

int
IMG_save_list_jf (IMG_list* img_list,char *filename)
{
  int total_class_nr=0;
  int total_img_nr=0;
  int img_nr;
  int dimension;
  int data_x;
  FILE *fp;
  
  dimension=img_list->img[0].x_size*img_list->img[0].y_size;
  
  for(img_nr=0;img_nr<img_list->total_nr;img_nr++)
    if(total_class_nr<img_list->img[img_nr].class)
      total_class_nr=img_list->img[img_nr].class;
  total_class_nr++; 
  
  
  if ((fp = fopen (filename, "w")) == NULL)
    {
      fprintf (stderr, "can't create file %s\n", filename);
      return (-1);
    }
  
  fprintf(fp,"%d %d\n",total_class_nr,dimension);
  
  for(img_nr=0;img_nr<img_list->total_nr;img_nr++)
    {
      if(img_list->img[img_nr].data_length[0]>0)
	{
	  fprintf(fp,"%d",img_list->img[img_nr].class);
	  for(data_x=0;
	      data_x<img_list->img[img_nr].data_length[0];data_x++)
	    fprintf(fp," %g",
		    img_list->img[img_nr].data_list[0][data_x]);
	  fprintf(fp,"\n");
	  total_img_nr++;
	}
    }
  fprintf (stdout,
	   "$IMG_save_list_jf(%s,%s);\n#saved %d images, x_size:%d y_size:%d\n",
	   img_list->name,filename, total_img_nr, img_list->img[0].x_size,
	   img_list->img[0].y_size);
  fclose (fp);
  return (0);
}


int
IMG_save_list_ejf (IMG_list* img_list,char *filename)
{
  int total_class_nr=0;
  int saved_img_depth=0;
  int img_nr;
  int img_depth;
  int data_x;
  FILE *fp;
  
  for(img_nr=0;img_nr<img_list->total_nr;img_nr++)
    if(total_class_nr<img_list->img[img_nr].class)
      total_class_nr=img_list->img[img_nr].class;
  total_class_nr++; 
  
  
  if ((fp = fopen (filename, "w")) == NULL)
    {
      fprintf (stderr, "can't create file %s\n", filename);
      return (-1);
    }
  
  for(img_depth=0;img_list->img[0].dis_mode[img_depth]!=-2;img_depth++)
    {
      if(img_list->img[0].data_length[img_depth]>0)
	{
	  saved_img_depth++;
	  fprintf(fp,"%d\n",total_class_nr);
  
	  for(img_nr=0;img_nr<img_list->total_nr;img_nr++)
	    {
		if(img_list->img[0].data_length[img_depth]==
		img_list->img[0].x_size*img_list->img[0].y_size)
	      		fprintf(fp,"%d %d %d",img_list->img[img_nr].class,
		      	img_list->img[img_nr].x_size,
		      	img_list->img[img_nr].y_size);
		else
			fprintf(fp,"%d %d 1",img_list->img[img_nr].class,
			img_list->img[img_nr].data_length[img_depth]);
											
   
	      for(data_x=0;
		  data_x<img_list->img[img_nr].data_length[img_depth];data_x++)
		fprintf(fp," %g",
			img_list->img[img_nr].data_list[img_depth][data_x]);
	      fprintf(fp,"\n");
	    }
	}
    }
  fprintf (stdout,
	   "$IMG_save_list_ejf(%s,%s);\n#saved %d images, img_depth:%d,  x_size:%d y_size:%d\n",
	   img_list->name,filename, img_list->total_nr, saved_img_depth, img_list->img[0].x_size,
	   img_list->img[0].y_size);
  fclose (fp);
  return (0);
}

int
IMG_save_list (IMG_list* img_list,char *filename)
{
  int img_nr;
  int warning_quadratic=0;
  int img_depth;

  if(strlen(filename)>=4 && strcmp(".ejf",&(filename[strlen(filename)-4]))==0)
    return(IMG_save_list_ejf(img_list,filename));
  else
    if(strlen(filename)>=3 && strcmp(".jf",&(filename[strlen(filename)-3]))==0) 
      {
	for (img_nr = 0; img_nr < img_list->total_nr; img_nr++)
	  {  
	    if((img_list->img[0].x_size != img_list->img[img_nr].y_size 
		|| img_list->img[0].x_size != img_list->img[img_nr].x_size) 
	       && !warning_quadratic)
	      {
		warning_quadratic=1;
		fprintf(stderr,"#Warning: You saved not quadratic images or images different sizes as a jf-file. You know what you are doing? Choose ejf-file and use imagepad to pad images, if you need quadratic images.\n");
	      }
	    if(warning_quadratic)
	      img_nr=img_list->total_nr;
	  }
	for(img_depth=0;img_list->img[0].dis_mode[img_depth]!=-2;img_depth++)
	  if(img_list->img[0].data_length[img_depth]==
	     img_list->img[0].x_size*img_list->img[0].y_size && img_depth>0)
	    {
	      fprintf(stderr,"#Warning: You saved only the first image depth of a multi depth image list. You know what you are doing? Choose ejf-file to save the whole image depth.\n"); 
	      break;
	    }
   
	return(IMG_save_list_jf(img_list,filename));
      }
    else
      {
	fprintf(stderr,"filename %s has to end with .jf or .ejf\n",filename);
	exit(-1);
      }
  return(-1);
}


int
IMG_print_img_info(IMG* img)
{
  int img_depth = 0;
  int x;
  
  while(img->dis_mode[img_depth] != -2)
    img_depth++;
  fprintf (stdout, "#   |  |-img_depth: %d\n", img_depth);
  for (x = 0; x < img_depth; x++)
    {
      fprintf (stdout, "#   |  |-dis_mode[%d]: %d scale: %f length: %d\n",
	       x, img->dis_mode[x], img->dis_scale[x],
	       img->data_length[x]);
    }
  
  return(0);
}

int
IMG_print_list_info (IMG_list * img_list)
{
  int img_depth = 0;
  int img_nr;
  int depth;
  int diff;

  fprintf (stdout, "#List->name: %s, ", img_list->name);
  fprintf (stdout, "->total_nr: %d\n", img_list->total_nr);
  fprintf (stdout, "# MAX +dim:%d ",IMG_get_max_img_dim_list(img_list));
  fprintf (stdout, "+x_size:%d ",IMG_get_max_img_x_size_list(img_list));
  fprintf (stdout, "+y_size:%d\n",IMG_get_max_img_y_size_list(img_list));
  fprintf (stdout, "# MIN +dim:%d ",IMG_get_min_img_dim_list(img_list));
  fprintf (stdout, "+x_size:%d ",IMG_get_min_img_x_size_list(img_list));
  fprintf (stdout, "+y_size:%d\n",IMG_get_min_img_y_size_list(img_list));
  fprintf (stdout, "#   |-img[0]\n");
  IMG_print_img_info(&(img_list->img[0])); 

  while(img_list->img[0].dis_mode[img_depth] != -2)
    img_depth++;
  for (img_nr = 1; img_nr < img_list->total_nr; img_nr++)
    {
      diff=0;
      for(depth=0;img_list->img[img_nr].dis_mode[depth] != -2 
	    && depth<img_depth; depth++)
	{
	  if(img_list->img[img_nr].dis_mode[depth]!=img_list->img[0].dis_mode[depth])
	    diff=1;
	  if(img_list->img[img_nr].dis_scale[depth]!=img_list->img[0].dis_scale[depth])
	    diff=1;
	}
      if(diff==1 || depth<img_depth || img_list->img[img_nr].dis_mode[depth] != -2)
	{
	  fprintf (stdout, "#   |-img[%d]\n",img_nr);
	  IMG_print_img_info(&(img_list->img[img_nr]));
	}
    }

  return 0;
}

int
IMG_copy (IMG A, IMG * B)
{
  int img_depth = 0;
  int x, j;

  while (A.dis_mode[img_depth] != -2)
    img_depth++;
  img_depth++;

  (*B).data_list = malloc (sizeof (double *) * img_depth);
  (*B).data_length = malloc (sizeof (int) * img_depth);
  (*B).dis_mode = malloc (sizeof (int) * img_depth);
  (*B).dis_scale = malloc (sizeof (double) * img_depth);
  if ((*B).data_list == NULL
      || (*B).data_length == NULL
      || (*B).dis_mode == NULL || (*B).dis_scale == NULL)
    {
      fprintf (stderr, "not enough memory to copy IMG\n");
      return (-1);
    }
  (*B).name = A.name;
  (*B).name = malloc(sizeof(char) * (strlen(A.name)+1));
  if ((*B).name == NULL)
    {
      fprintf (stderr, "not enough memory to copy IMG\n");
      exit (-1);
    }
  strcpy((*B).name,A.name);
  (*B).x_size = A.x_size;
  (*B).y_size = A.y_size;
  for (x = 0; x < img_depth - 1; x++)
    {
      (*B).data_list[x] = malloc (A.data_length[x] * sizeof (double));
      if ((*B).data_list[x] == NULL)
	{
	  fprintf (stderr, "not enough memory to copy IMG\n");
	  return (-1);
	}
      for (j = 0; j < A.data_length[x]; j++)
	(*B).data_list[x][j] = A.data_list[x][j];
      (*B).data_length[x] = A.data_length[x];
      (*B).dis_mode[x] = A.dis_mode[x];
      (*B).dis_scale[x] = A.dis_scale[x];
      (*B).class = A.class;
      (*B).beamsize = A.beamsize;
      (*B).dis_limit = A.dis_limit;
      (*B).warprange = A.warprange;
    }
  (*B).data_list[img_depth - 1] = NULL;
  (*B).dis_mode[img_depth - 1] = -2;
  return (0);
}

IMG_list *
IMG_copy_list (IMG_list * source_list)
{
  IMG_list *dest_list;
  int i;

  dest_list = malloc (sizeof (IMG_list));
  if (dest_list == NULL)
    {
      fprintf (stderr, "not enough memory to copy IMG_list %s\n",
	       source_list->name);
      exit (-1);
    }
  dest_list->img = malloc (source_list->total_nr * sizeof (IMG));
  if (dest_list->img == NULL
      || NULL == (dest_list->name=(char*) malloc((strlen(source_list->name)+1)*sizeof(char))))
    {
      fprintf (stderr, "not enough memory to copy IMG_list %s\n",
	       source_list->name);
      exit (-1);
    }

  dest_list->total_nr = source_list->total_nr;

  strcpy(dest_list->name, source_list->name);

  for (i = 0; i < dest_list->total_nr; i++)
    {
      if (IMG_copy (source_list->img[i], &(dest_list->img[i])) != 0)
	{
	  fprintf (stderr, "not enough memory to copy IMG_list %s\n",
		   source_list->name);
	  exit (-1);
	}

    }
  fprintf (stdout, "$IMG_copy_list(%s)\n", source_list->name);
  return (dest_list);
}

IMG_list*
IMG_create_new_list(int img_nr, int img_depth)
{
  IMG_list *dest_list; 
  int i,j;
  char* new_list_name="new_img_list";

  dest_list = malloc (sizeof (IMG_list));
  if (dest_list == NULL)
    {
      fprintf (stderr,
	       "not enough memory to create a new IMG_list\n");
      exit (-1);
    }
  dest_list->img = malloc (img_nr * sizeof (IMG));
  if (dest_list->img == NULL
      || NULL == (dest_list->name=(char*) malloc((strlen(new_list_name)+1)*sizeof(char))))
    {
      fprintf (stderr,
	       "not enough memory to create a new IMG_list\n");
      exit (-1);
    }   
  dest_list->total_nr = img_nr;
  strcpy(dest_list->name,new_list_name);
  for (i = 0; i < dest_list->total_nr; i++)
    {
      dest_list->img[i].data_list = malloc (sizeof (double *) * img_depth);
      dest_list->img[i].data_length = malloc (sizeof (int) * img_depth);
      dest_list->img[i].dis_mode = malloc (sizeof (int) * img_depth);
      dest_list->img[i].dis_scale = malloc (sizeof (double) * img_depth);
      if (dest_list->img[i].data_list == NULL
	  || dest_list->img[i].data_length == NULL
	  || dest_list->img[i].dis_mode == NULL
	  || dest_list->img[i].dis_scale == NULL)
	{
	  fprintf (stderr,
		   "not enough memory to create an emptylist IMG_list %s\n",
		   dest_list->name);
	  exit (-1);
	}
      dest_list->img[i].x_size = 0;
      dest_list->img[i].y_size = 0;
      dest_list->img[i].class = -1;
      for (j = 0; j < (img_depth - 1); j++)
	{
	  dest_list->img[i].data_length[j] = 0;
	  dest_list->img[i].data_list[j] = NULL;
	  dest_list->img[i].dis_mode[j] = -1;
	}
      dest_list->img[i].data_list[img_depth - 1] = NULL;
      dest_list->img[i].dis_mode[img_depth - 1] = -2;
    }
  fprintf (stdout, "$IMG_create_new_list(%d,%d)\n", img_nr,
	   img_depth);
  return (dest_list);
}

IMG_list *
IMG_create_empty_list (IMG_list * source_list, int img_depth)
{
  IMG_list *dest_list;
  int i, j;
  img_depth++;
  dest_list = malloc (sizeof (IMG_list));
  if (dest_list == NULL)
    {
      fprintf (stderr,
	       "not enough memory to create an emptylist IMG_list %s\n",
	       source_list->name);
      exit (-1);
    }
  dest_list->img = malloc (source_list->total_nr * sizeof (IMG));
  if (dest_list->img == NULL
      || NULL == (dest_list->name=(char*) malloc((strlen(source_list->name)+1)*sizeof(char))))
    {
      fprintf (stderr,
	       "not enough memory to create an emptylist IMG_list %s\n",
	       source_list->name);
      exit (-1);
    }

  dest_list->total_nr = source_list->total_nr;

  strcpy(dest_list->name, source_list->name);

  for (i = 0; i < dest_list->total_nr; i++)
    {
      dest_list->img[i].data_list = malloc (sizeof (double *) * img_depth);
      dest_list->img[i].data_length = malloc (sizeof (int) * img_depth);
      dest_list->img[i].dis_mode = malloc (sizeof (int) * img_depth);
      dest_list->img[i].dis_scale = malloc (sizeof (double) * img_depth);
      if (dest_list->img[i].data_list == NULL
	  || dest_list->img[i].data_length == NULL
	  || dest_list->img[i].dis_mode == NULL
	  || dest_list->img[i].dis_scale == NULL)
	{
	  fprintf (stderr,
		   "not enough memory to create an emptylist IMG_list %s\n",
		   source_list->name);
	  exit (-1);
	}
      dest_list->img[i].name = malloc(sizeof(char) * (strlen(source_list->img[i].name)+1));
      if (dest_list->img[i].name == NULL)
	{
	  fprintf (stderr,
		   "not enough memory to create an emptylist IMG_list %s\n",
		   source_list->name);
	  exit (-1);
	}
      strcpy(dest_list->img[i].name,source_list->img[i].name);
      dest_list->img[i].x_size = source_list->img[i].x_size;
      dest_list->img[i].y_size = source_list->img[i].y_size;
      dest_list->img[i].class = source_list->img[i].class;
      for (j = 0; j < (img_depth - 1); j++)
	{
	  dest_list->img[i].data_length[j] = 0;
	  dest_list->img[i].data_list[j] = NULL;
	  dest_list->img[i].dis_mode[j] = -1;
	}
      dest_list->img[i].data_list[img_depth - 1] = NULL;
      dest_list->img[i].dis_mode[img_depth - 1] = -2;
    }
  fprintf (stdout, "$IMG_create_empty_list(%s,%d)\n", source_list->name,
	   img_depth);
  return (dest_list);
}

int
IMG_insert_in_list (IMG_list * source_list, IMG_list * dest_list,
		    int position)
{
  int i, j;
  int x = 0;
  int start_pos;

  start_pos = position;

  if (dest_list->total_nr != source_list->total_nr)
    {
      fprintf (stderr, "list->total_nr is not equal %s %s\n",
	       source_list->name, dest_list->name);
      exit (-1);
    }
  // printf("source:%d des:%d\n",source_list->img[0].dis_mode[x],dest_list->img[0].dis_mode[position]);
  while (source_list->img[0].dis_mode[x] != -2
	 && dest_list->img[0].dis_mode[position] != -2)
    {
      //   printf("bla");
      for (i = 0; i < dest_list->total_nr; i++)
	{

	  if (dest_list->img[i].data_list[position] != NULL)
	    free (dest_list->img[i].data_list[position]);
	  dest_list->img[i].data_list[position] =
	    malloc (source_list->img[i].data_length[x] * sizeof (double));
	  if (dest_list->img[i].data_list[position] == NULL)
	    {
	      fprintf (stderr, "not enough memory to copy IMG_list %s\n",
		       source_list->name);
	      exit (-1);
	    }
	  for (j = 0; j < source_list->img[i].data_length[x]; j++)
	    dest_list->img[i].data_list[position][j] =
	      source_list->img[i].data_list[x][j];
	  dest_list->img[i].data_length[position] =
	    source_list->img[i].data_length[x];
	  dest_list->img[i].dis_mode[position] =
	    source_list->img[i].dis_mode[x];
	  dest_list->img[i].dis_scale[position] =
	    source_list->img[i].dis_scale[x];
	}
      position++;
      x++;
    }

  fprintf (stdout, "$IMG_insert_in_list(%s,%s,%d) //end_pos:%d\n",
	   source_list->name, dest_list->name, start_pos, position);
  return (0);
}

int
IMG_move_in_list(IMG_list* source_list,int source_pos, IMG_list* dest_list,int dest_pos)
{
  int i;

  if (dest_list->total_nr != source_list->total_nr)
    {
      fprintf (stderr, "list->total_nr is not equal %s %s\n",
	       source_list->name, dest_list->name);
      exit (-1);
    }

  for (i = 0; i < dest_list->total_nr; i++)
    {

      if (dest_list->img[i].data_list[dest_pos] != NULL)
	free (dest_list->img[i].data_list[dest_pos]);
      dest_list->img[i].data_list[dest_pos]=
	source_list->img[i].data_list[source_pos];      
      source_list->img[i].data_list[source_pos]=NULL;
      dest_list->img[i].data_length[dest_pos] =
	source_list->img[i].data_length[source_pos];
      source_list->img[i].data_length[source_pos]=0;
      dest_list->img[i].dis_mode[dest_pos] =
	source_list->img[i].dis_mode[source_pos];
      source_list->img[i].dis_mode[source_pos]=-1;
      dest_list->img[i].dis_scale[dest_pos] =
	source_list->img[i].dis_scale[source_pos];
    }

  fprintf (stdout, "$IMG_move_in_list(%s,%d,%s,%d)\n",
	   source_list->name, source_pos,
	   dest_list->name, dest_pos);  
  return(0);
}

int
IMG_add_dataspace_in_list(IMG_list* img_list, int img_depth_pos, int data_dim)
{
  int img_nr;
  
  for (img_nr = 0; img_nr < img_list->total_nr; img_nr++)
    {
      if (img_list->img[img_nr].data_list[img_depth_pos] != NULL)
	free (img_list->img[img_nr].data_list[img_depth_pos]);
      
      img_list->img[img_nr].data_list[img_depth_pos]=
	malloc(data_dim*sizeof(double));
      if(img_list->img[img_nr].data_list[img_depth_pos]==NULL)
	{
	  fprintf (stderr,
		   "not enough memory to create an dataspace in IMG_list %s\n",
		   img_list->name);
	  exit (-1);
	}
      img_list->img[img_nr].data_length[img_depth_pos]=data_dim;
    }

  fprintf (stdout, "$IMG_add_dataspace_in_list(%s,%d,%d)\n",
	   img_list->name, img_depth_pos, data_dim);  
  return(0);
}

int
IMG_free (IMG * img)
{
  int img_depth = 0;
  int x;

  while ((*img).dis_mode[img_depth] != -2)
    img_depth++;
  img_depth++;

  for (x = 0; x < img_depth; x++)
    {
      if ((*img).data_list[x] != NULL)
	free ((*img).data_list[x]);
    }

  if( (*img).name!=NULL )
    free((*img).name);
  free ((*img).data_list);
  free ((*img).data_length);
  free ((*img).dis_mode);
  free ((*img).dis_scale);

  return (0);
}

int
IMG_free_list (IMG_list * img_list)
{
  int i;

  for (i = 0; i < img_list->total_nr; i++)
    IMG_free (&(img_list->img[i]));
  
  free (img_list->img);

  fprintf (stdout, "$IMG_free_list(%s)\n", img_list->name);
  free (img_list->name);
  free (img_list);
    
  return (0);
}

int
IMG_list_set_beamsize (IMG_list * img_list, int beamsize)
{
  int x;

  for (x = 0; x < img_list->total_nr; x++)
    img_list->img[x].beamsize = beamsize;

  fprintf (stdout, "$IMG_list_set_beamsize(%s,%d)\n", img_list->name,
	   beamsize);

  return (0);
}

int
IMG_list_set_dis_limit (IMG_list * img_list, double dis_limit)
{
  int x;

  for (x = 0; x < img_list->total_nr; x++)
    img_list->img[x].dis_limit = dis_limit;

  fprintf (stdout, "$IMG_list_set_dis_limit(%s,%f)\n", img_list->name,
	   dis_limit);

  return (0);
}

int
IMG_list_set_warprange (IMG_list * img_list, int warprange)
{
  int x;

  for (x = 0; x < img_list->total_nr; x++)
    img_list->img[x].warprange = warprange;

  fprintf (stdout, "$IMG_list_set_warprange(%s,%d)\n", img_list->name,
	   warprange);

  return (0);
}

int
IMG_list_set_dis_mode (IMG_list * img_list, int field, int dis_mode)
{
  int x;

  for (x = 0; x < img_list->total_nr; x++)
    img_list->img[x].dis_mode[field] = dis_mode;

  fprintf (stdout, "$IMG_list_set_dis_mode(%s,%d,%d)\n", img_list->name,
	   field, dis_mode);

  return (0);
}

int
IMG_list_set_dis_scale (IMG_list * img_list, int field, double dis_scale)
{
  int x;

  for (x = 0; x < img_list->total_nr; x++)
    img_list->img[x].dis_scale[field] = dis_scale;

  fprintf (stdout, "$IMG_list_set_dis_scale(%s,%d,%f)\n", img_list->name,
	   field, dis_scale);

  return (0);
}

int
IMG_get_max_dim_list(IMG_list* img_list)
{
	int img_nr;
	int max_dim = 0;
  for (img_nr = 0; img_nr < img_list->total_nr; img_nr++)
	if(max_dim<(img_list->img[img_nr].x_size*img_list->img[img_nr].y_size))
	  max_dim=(img_list->img[img_nr].x_size*img_list->img[img_nr].y_size);

	return(max_dim);
}

int
IMG_get_min_max_list (IMG_list * img_list, int field, double *min,
		      double *max)
{
  int img_nr;
  int x;

  for (img_nr = 0; img_nr < img_list->total_nr; img_nr++)
    {
      for (x = 0; x < img_list->img[img_nr].data_length[field]; x++)
	{
	  if (*max < img_list->img[img_nr].data_list[field][x])
	    *max = img_list->img[img_nr].data_list[field][x];
	  if (*min > img_list->img[img_nr].data_list[field][x])
	    *min = img_list->img[img_nr].data_list[field][x];
	}
    }
  fprintf (stdout, "$IMG_get_min_max_list(%s,%d,%f,%f);\n", img_list->name,
	   field, *min, *max);
  return (0);
}

int
IMG_norm_list (IMG_list * img_list, int field, double min, double max,
	       double new_min, double new_max)
{
  int img_nr;
  int x;

  for (img_nr = 0; img_nr < img_list->total_nr; img_nr++)
    for (x = 0; x < img_list->img[img_nr].data_length[field]; x++)
      img_list->img[img_nr].data_list[field][x] =
	(((img_list->img[img_nr].data_list[field][x] - min) / (max -
							       min)) *
	 (new_max - new_min)) + new_min;

  fprintf (stdout, "$IMG_norm_list(%s,%d,%f,%f,%f,%f);\n", img_list->name,
	   field, min, max, new_min, new_max);
  return (0);
}

int
IMG_set_data_in_list (IMG_list * img_list, int field, double value)
{
  int img_nr;
  int x;

  for (img_nr = 0; img_nr < img_list->total_nr; img_nr++)
    for (x = 0; x < img_list->img[img_nr].data_length[field]; x++)
      img_list->img[img_nr].data_list[field][x] = value;

  fprintf (stdout, "$IMG_set_data_in_list(%s,%d,%f);\n", img_list->name,
	   field, value);
  return (0);
}

int
IMG_add_border (IMG * img, int l_b, int r_b, int u_b, int d_b,
		double border_color)
{
  int x, y, b, new_x_size, new_y_size;
  double *temp_data;
  int img_depth;

  new_x_size = img->x_size + l_b + r_b;
  new_y_size = img->y_size + u_b + d_b;

  for(img_depth=0;img->dis_mode[img_depth]!=-2;img_depth++)
    {
      if(img->data_length[img_depth]==
	 img->x_size*img->y_size)
	{
	  temp_data = malloc (sizeof (double) * (new_x_size) * (new_y_size));
	  if (temp_data == NULL)
	    {
	      fprintf (stderr, "not enough memory for IMG_add_border\n");
	      exit (0);
	    }

	  for (x = 0; x < new_x_size; x++)
	    {
	      for (b = 0; b < u_b; b++)
		temp_data[((b) * new_x_size) + x] = border_color;
	      for (b = 0; b < d_b; b++)
		temp_data[((new_y_size - 1 - b) * new_x_size) + x] = border_color;
	    }
	  for (y = 0; y < new_y_size; y++)
	    {
	      for (b = 0; b < l_b; b++)
		temp_data[(y * new_x_size) + b] = border_color;
	      for (b = 0; b < r_b; b++)
		temp_data[((y + 1) * new_x_size) - 1 - b] = border_color;
	    }

	  for (y = 0; y < img->y_size; y++)
	    for (x = 0; x < img->x_size; x++)
	      temp_data[((y + u_b) * (new_x_size)) + (x + l_b)] =
		img->data_list[img_depth][(y * img->x_size) + x];
	  free ((*img).data_list[img_depth]);
	  img->data_list[img_depth] = temp_data;
	  img->data_length[img_depth] = new_x_size * new_y_size;	  
	}
    }

  img->x_size = new_x_size;
  img->y_size = new_y_size;
  return (0);
}

int
IMG_add_border_list (IMG_list * img_list, int l_b, int r_b, int u_b, int d_b,
		     double border_color)
{
  int img_nr;

  for (img_nr = 0; img_nr < img_list->total_nr; img_nr++)
    IMG_add_border (&(img_list->img[img_nr]), l_b, r_b, u_b, d_b,
		    border_color);

  fprintf (stdout,
	   "$IMG_add_border_list(%s,%d,%d,%d,%d,%f);\n",
	   img_list->name, l_b, r_b, u_b, d_b, border_color);
  return (0);
}

IMG_list *
IMG_shift3x3_list (IMG_list * source_list, double bg_color)
{
  IMG_list *dest_list;
  int i, j, x, y, v, w;

  dest_list = malloc (sizeof (IMG_list));
  if (dest_list == NULL)
    {
      fprintf (stderr,
	       "not enough memory to create an emptylist IMG_list %s\n",
	       source_list->name);
      exit (-1);
    }
  dest_list->img = malloc (source_list->total_nr * sizeof (IMG) * 9);
  if (dest_list->img == NULL
      || NULL == (dest_list->name=(char*) malloc((strlen(source_list->name)+1)*sizeof(char))))
    {
      fprintf (stderr, "not enough memory to shift3x3_list %s\n",
	       source_list->name);
      exit (-1);
    }
  dest_list->total_nr = source_list->total_nr * 9;

  strcpy(dest_list->name, source_list->name);

  j = 0;
  for (i = 0; i < source_list->total_nr; i++)
    {
      for (x = -1; x < 2; x++)
	{
	  for (y = -1; y < 2; y++)
	    {
	      dest_list->img[j].data_list = malloc (sizeof (double *) * 2);
	      dest_list->img[j].data_length = malloc (sizeof (int) * 2);
	      dest_list->img[j].dis_mode = malloc (sizeof (int) * 2);
	      dest_list->img[j].dis_scale = malloc (sizeof (double) * 2);
	      if (dest_list->img[j].data_list == NULL
		  || dest_list->img[j].data_length == NULL
		  || dest_list->img[j].dis_mode == NULL
		  || dest_list->img[j].dis_scale == NULL)
		{
		  fprintf (stderr, "not enough memory to shift3x3_list %s\n",
			   source_list->name);
		  exit (-1);
		}
	      dest_list->img[j].name = malloc(sizeof(char) * (strlen(source_list->img[i].name)+1));
	      if (dest_list->img[j].name == NULL)
		{
		  fprintf (stderr, "not enough memory to shift3x3_list %s\n",
			   source_list->name);
		  exit (-1);
		}
	      strcpy(dest_list->img[j].name,source_list->img[i].name);
	      dest_list->img[j].x_size = source_list->img[i].x_size;
	      dest_list->img[j].y_size = source_list->img[i].y_size;
	      dest_list->img[j].class = source_list->img[i].class;

	      dest_list->img[j].data_list[1] = NULL;
	      dest_list->img[j].dis_mode[1] = -2;
	      dest_list->img[j].dis_mode[0] = -1;
	      dest_list->img[j].data_length[0] =
		dest_list->img[i].x_size * dest_list->img[i].y_size;
	      dest_list->img[j].data_list[0] =
		malloc (sizeof (double) * dest_list->img[j].data_length[0]);
	      if (dest_list->img[j].data_list[0] == NULL)
		{
		  fprintf (stderr, "not enough memory to shift3x3_list %s\n",
			   source_list->name);
		  exit (-1);
		}
	      for (w = 0; w < dest_list->img[j].y_size; w++)
		{
		  for (v = 0; v < dest_list->img[j].x_size; v++)
		    {
		      if ((v + x) < 0 || (v + x) >= dest_list->img[j].x_size
			  || (w + y) < 0
			  || (w + y) >= dest_list->img[j].y_size)
			dest_list->img[j].data_list[0][v +
						       (w *
							dest_list->img[j].
							x_size)] = bg_color;
		      else
			dest_list->img[j].data_list[0][v +
						       (w *
							dest_list->img[j].
							x_size)] =
			  source_list->img[i].data_list[0][v + x +
							   ((w +
							     y) *
							    dest_list->img[j].
							    x_size)];

		    }
		}
	      j++;
	    }
	}
    }
  fprintf (stdout, "$IMG_shift3x3_list()\n");
  return (dest_list);
}

int
IMG_flip (IMG * img, int flip_mode)
{
  int x, y;
  double temp;
  double *temp_data;

  if (flip_mode & IMG_VERTICAL)
    {
      for (y = 0; y < img->y_size; y++)
	for (x = 0; x < (img->x_size / 2); x++)
	  {
	    temp = img->data_list[0][(y * img->x_size) + x];
	    img->data_list[0][(y * img->x_size) + x] =
	      img->data_list[0][(y * img->x_size) + img->x_size - 1 - x];
	    img->data_list[0][(y * img->x_size) + img->x_size - 1 - x] = temp;
	  }
    }

  if (flip_mode & IMG_HORIZONTAL)
    {
      for (x = 0; x < img->x_size; x++)
	for (y = 0; y < (img->y_size / 2); y++)
	  {
	    temp = img->data_list[0][(y * img->x_size) + x];
	    img->data_list[0][(y * img->x_size) + x] =
	      img->data_list[0][((img->y_size - 1 - y) * img->x_size) + x];
	    img->data_list[0][((img->y_size - 1 - y) * img->x_size) + x] =
	      temp;
	  }
    }

  if (flip_mode & IMG_ROTATE)
    {
      temp_data = malloc (sizeof (double) * img->x_size * img->y_size);
      if (temp_data == NULL)
	{
	  fprintf (stderr, "not enough memory for IMG_ROTATE\n");
	  exit (0);
	}

      for (y = 0; y < img->y_size; y++)
	for (x = 0; x < img->x_size; x++)
	  temp_data[(x * img->x_size) + y] =
	    img->data_list[0][((img->y_size - 1 - y) * img->x_size) + x];

      for (x = 0; x < img->data_length[0]; x++)
	img->data_list[0][x] = temp_data[x];

      free (temp_data);
    }
  return 0;
}

int
IMG_flip_list (IMG_list * img_list, int flip_mode)
{
  int img_nr;

  for (img_nr = 0; img_nr < img_list->total_nr; img_nr++)
    IMG_flip (&(img_list->img[img_nr]), flip_mode);

  fprintf (stdout, "$IMG_flip_list(%s,", img_list->name);
  if (flip_mode & IMG_VERTICAL)
    fprintf (stdout, "IMG_VERTICAL ");
  if (flip_mode & IMG_HORIZONTAL)
    fprintf (stdout, "IMG_HORIZONTAL ");
  if (flip_mode & IMG_ROTATE)
    fprintf (stdout, "IMG_ROTATE ");
  fprintf (stdout, ");\n");
  return (0);
}

IMG_list*
IMG_create_mean_img_list(IMG_list* img_list_src,double W2D_function(IMG,IMG,int*),int iter_total)
{
  IMG_list* img_list_mean;
  IMG_list* img_list_temp;
  int* img_counter=NULL;
  int img_nr_src;
  int img_nr_mean;
  int data_x;
  int img_depth;
  int* warp_data;
  int* warp_counter;
  int iter_nr;
  int iter_stop;

  warp_data=malloc (sizeof (int) * IMG_get_max_dim_list(img_list_src) * 2);
  warp_counter=malloc (sizeof (int) * IMG_get_max_dim_list(img_list_src));
  img_list_mean = malloc (sizeof (IMG_list));
  
  if (warp_data == NULL || warp_counter == NULL || img_list_mean == NULL 
      || NULL == (img_list_mean->name=(char*) malloc((strlen(img_list_src->name)+1)*sizeof(char))))
    {
      fprintf (stderr, "not enough memory to create mean IMG_list %s\n",
	       img_list_src->name);
      exit (-1);
    }


  img_list_mean->img = NULL;
  img_list_mean->total_nr = 0;
  strcpy(img_list_mean->name, img_list_src->name);  

  for(img_nr_src=0;img_nr_src<img_list_src->total_nr;img_nr_src++)
    {
      img_nr_mean=0;
      while(img_nr_mean<img_list_mean->total_nr
	    &&  img_list_mean->img[img_nr_mean].class!=img_list_src->img[img_nr_src].class)
	img_nr_mean++;
      
      if(img_nr_mean==img_list_mean->total_nr)
	{
	  img_list_mean->total_nr++;
	  img_list_mean->img=realloc(img_list_mean->img,
				  img_list_mean->total_nr*sizeof(IMG));
	  if (img_list_mean->img == NULL)
	    {
	      fprintf (stderr, "not enough memory to create mean IMG_list %s\n",
		       img_list_mean->name);
	      exit (-1);
	    }
	  if(IMG_copy(img_list_src->img[img_nr_src],&(img_list_mean->img[img_nr_mean])))
	    exit(-1);
	  img_counter=realloc(img_counter,
			      img_list_mean->total_nr*sizeof(int));
	  if (img_counter == NULL)
	    {
	      fprintf (stderr, "not enough memory to create mean IMG_list %s\n",
		       img_list_mean->name);
	      exit (-1);
	    }
	  img_counter[img_nr_mean]=1;
	}
      else
	{
	  img_counter[img_nr_mean]++;
	  for(img_depth=0;img_list_src->img[img_nr_src].dis_mode[img_depth]!=-2;img_depth++)
	    for(data_x=0;data_x<img_list_src->img[img_nr_src].data_length[img_depth];data_x++)
	      img_list_mean->img[img_nr_mean].data_list[img_depth][data_x]+=
		img_list_src->img[img_nr_src].data_list[img_depth][data_x];
	}
    }


  for(img_nr_mean=0;img_nr_mean<img_list_mean->total_nr;img_nr_mean++)
    for(img_depth=0;img_list_mean->img[img_nr_mean].dis_mode[img_depth]!=-2;img_depth++)
      for(data_x=0;data_x<img_list_mean->img[img_nr_mean].data_length[img_depth];data_x++)
	img_list_mean->img[img_nr_mean].data_list[img_depth][data_x]/=
	  img_counter[img_nr_mean];

  fprintf(stdout,"#calculated mean class images\n");

  /*
  for(img_depth=0;merge_list->img[0].dis_mode[img_depth]!=-2;img_depth++)
    {
      min = DBL_MAX;
      max = -DBL_MAX;
      IMG_get_min_max_list (merge_list, img_depth, &min, &max);
      IMG_norm_list (merge_list, img_depth, min, max, 0.0, 1.0);
    }
  */

  for(iter_nr=0;iter_nr<iter_total;iter_nr++)
    {

  img_list_temp=IMG_copy_list(img_list_mean);
  for(img_nr_mean=0;img_nr_mean<img_list_temp->total_nr;img_nr_mean++)
    for(img_depth=0;img_list_temp->img[img_nr_mean].dis_mode[img_depth]!=-2;img_depth++)
      for(data_x=0;data_x<img_list_temp->img[img_nr_mean].data_length[img_depth];data_x++)
	img_list_temp->img[img_nr_mean].data_list[img_depth][data_x]=0.0;


  for(img_nr_mean=0;img_nr_mean<img_list_temp->total_nr;img_nr_mean++)
    {
      for(data_x=0;data_x<img_list_temp->img[img_nr_mean].x_size
	    *img_list_temp->img[img_nr_mean].y_size;data_x++)
	warp_counter[data_x]=0;

      for(img_nr_src=0;img_nr_src<img_list_src->total_nr;img_nr_src++)
	{
	  if(img_list_mean->img[img_nr_mean].class==img_list_src->img[img_nr_src].class)
	    {
	      W2D_function(img_list_src->img[img_nr_src],img_list_mean->img[img_nr_mean],warp_data);
	      
	      for(img_depth=0;img_list_temp->img[img_nr_mean].dis_mode[img_depth]!=-2;img_depth++)
		for(data_x=0;data_x<img_list_temp->img[img_nr_mean].data_length[img_depth];data_x++)
		  {
		    img_list_temp->img[img_nr_mean].data_list[img_depth]
		      [(warp_data[(data_x*2)+1]*img_list_temp->img[img_nr_mean].x_size)
		       +warp_data[(data_x*2)]]+=
		      img_list_src->img[img_nr_src].data_list[img_depth][data_x];
		    if(img_depth==0)
		      warp_counter[(warp_data[(data_x*2)+1]*img_list_temp->img[img_nr_mean].x_size)
				   +warp_data[(data_x*2)]]++;
		  }
	    }
	}

      for(img_depth=0;img_list_temp->img[img_nr_mean].dis_mode[img_depth]!=-2;img_depth++)	  
	for(data_x=0;data_x<img_list_temp->img[img_nr_mean].data_length[img_depth];data_x++)
	  {
	    if(warp_counter[data_x]!=0)
	      img_list_temp->img[img_nr_mean].data_list[img_depth][data_x]/=	      
		warp_counter[data_x];
	  }

    }

  fprintf(stdout,"#Iteration: %d calculated mean class images from warped images\n",iter_nr+1);

  iter_stop=1;
  for(img_nr_mean=0;img_nr_mean<img_list_temp->total_nr;img_nr_mean++)
    for(img_depth=0;img_list_temp->img[img_nr_mean].dis_mode[img_depth]!=-2;img_depth++)
      for(data_x=0;(data_x<img_list_temp->img[img_nr_mean].data_length[img_depth]
		    && iter_stop==1);data_x++)
	if(img_list_temp->img[img_nr_mean].data_list[img_depth][data_x]
	   !=img_list_mean->img[img_nr_mean].data_list[img_depth][data_x])
	  iter_stop=0;
  
  if(iter_stop==1)
    {
      fprintf(stdout,"#Stop iteration cycle, because the new created img_list hasn't changed\n");
      iter_nr=iter_total;
    }

  IMG_free_list(img_list_mean); 
  img_list_mean=img_list_temp;

  /*  for(img_depth=0;merge_list->img[0].dis_mode[img_depth]!=-2;img_depth++)
    {
      min = DBL_MAX;
      max = -DBL_MAX;
      IMG_get_min_max_list (merge_list, img_depth, &min, &max);
      IMG_norm_list (merge_list, img_depth, min, max, 0.0, 1.0);
    }
  */
    }

  free(img_counter);
  free(warp_data);
  free(warp_counter);
  return(img_list_mean);
}

int IMG_add_position_x_list(IMG_list *img_list) {
  int img_nr;
  
  for (img_nr = 0; img_nr < img_list->total_nr; img_nr++)
    IMG_add_position_x_img (&(img_list->img[img_nr]));

  return 0;
}

void IMG_add_position_x_img(IMG *img) {
  int y,x; 
  for (y = 0; y < img->y_size; y++) {
    for (x = 0; x < img->x_size; x++) {
      img->data_list[0][y*img->x_size+x]=x;
    }
  }
}

int IMG_add_position_y_list(IMG_list *img_list) {
  int img_nr;
  
  for (img_nr = 0; img_nr < img_list->total_nr; img_nr++)
    IMG_add_position_y_img (&(img_list->img[img_nr]));

  return 0;
}

void IMG_add_position_y_img(IMG *img) {
  int y,x; 
  for (y = 0; y < img->y_size; y++) {
    for (x = 0; x < img->x_size; x++) {
      img->data_list[0][y*img->x_size+x]=y;
    }
  }
}




int
IMG_filter_img (IMG * img, int filter_mode)
{
  double SobelVplus[] = { +0.0, -1.0, +0.0, +1.0, +0.0,
			  -0.5, -2.0, +0.0, +2.0, +0.5,
			  +0.0, -1.0, +0.0, +1.0, +0.0
  };
  
  double SobelHplus[] = { +0.0, -0.5, +0.0,
			  -1.0, -2.0, -1.0,
			  +0.0, +0.0, +0.0,
			  +1.0, +2.0, +1.0,
			  +0.0, +0.5, +0.0
  };
  
  double SobelDplus[] = { +0.0, +0.0, +0.0, +0.0, -0.5,
			  +0.0, +0.0, -1.0, -2.0, +0.0,
			  +0.0, +1.0, +0.0, -1.0, +0.0,
			  +0.0, +2.0, +1.0, +0.0, +0.0,
			  +0.5, +0.0, +0.0, +0.0, +0.0
  };

  double SobelV[] = { -1.0, +0.0, +1.0,
		      -2.0, +0.0, +2.0,
		      -1.0, +0.0, +1.0
  };
  
  double SobelH[] = { -1.0, -2.0, -1.0,
		      +0.0, +0.0, +0.0,
		      +1.0, +2.0, +1.0
  };
  
  double SobelD[] = { +0.0, -1.0, -2.0,
		      +1.0, +0.0, -1.0,
		      +2.0, +1.0, +0.0
  };

  double SobelVjaehne[] = { -3.0, +0.0, +3.0,
			    -10.0, +0.0, +10.0,
			    -3.0, +0.0, +3.0
  };
  
  double SobelHjaehne[] = { -3.0, -10.0, -3.0,
			    +0.0, +0.0, +0.0,
			    +3.0, +10.0, +3.0
  };
  
  double SobelDjaehne[] = { +0.0, -10.0, -3.0,
			    +3.0, +0.0, -10.0,
			    +10.0, +3.0, +0.0
  };

  double Laplace[] = { +0.0, +1.0, +0.0,
		       +1.0, -4.0, +1.0,
		       +0.0, +1.0, +0.0
  };

  double *filter=NULL;
  int f_x_size=0;
  int f_y_size=0;
  int x, y, f_x, f_y, f_ox, f_oy, x_temp, y_temp;
  double f_temp;
  double *temp_data;
  double f_div=0.0;

  temp_data = malloc (sizeof (double) * img->x_size * img->y_size);
  if (temp_data == NULL)
    {
      fprintf (stderr, "not enough memory for IMG_filter_img\n");
      exit (0);
    }

  switch (filter_mode)
    {
    case IMG_sobelhplus:
      f_x_size = 3;
      f_y_size = 5;
      filter = SobelHplus;
      f_div = 4.5;
      break;
    case IMG_sobelvplus:
      f_x_size = 5;
      f_y_size = 3;
      filter = SobelVplus;
      f_div = 4.5;
      break;
    case IMG_sobeldplus:
      f_x_size = 5;
      f_y_size = 5;
      filter = SobelDplus;
      f_div = 4.5;
      break;
   case IMG_sobelhjaehne:
      f_x_size = 3;
      f_y_size = 3;
      filter = SobelHjaehne;
      f_div = 16;
      break;
    case IMG_sobelvjaehne:
      f_x_size = 3;
      f_y_size = 3;
      filter = SobelVjaehne;
      f_div = 16;
      break;
    case IMG_sobeldjaehne:
      f_x_size = 3;
      f_y_size = 3;
      filter = SobelDjaehne;
      f_div = 16;
      break;
   case IMG_sobelh:
      f_x_size = 3;
      f_y_size = 3;
      filter = SobelH;
      f_div = 4;
      break;
    case IMG_sobelv:
      f_x_size = 3;
      f_y_size = 3;
      filter = SobelV;
      f_div = 4;
      break;
    case IMG_sobeld:
      f_x_size = 3;
      f_y_size = 3;
      filter = SobelD;
      f_div = 4;
      break;
    case IMG_laplace:
      f_x_size = 3;
      f_y_size = 3;
      filter = Laplace;
      f_div = 4;
      break;
    }

  f_ox = -(f_x_size - 1) / 2;
  f_oy = -(f_y_size - 1) / 2;

  for (y = 0; y < img->y_size; y++)
    for (x = 0; x < img->x_size; x++)
      {
	f_temp = 0;
	for (f_y = 0; f_y < f_y_size; f_y++)
	  for (f_x = 0; f_x < f_x_size; f_x++)
	    {
	      x_temp = f_x + f_ox + x;
	      y_temp = f_y + f_oy + y;
	      if (x_temp >= 0 && x_temp < img->x_size && y_temp >= 0
		  && y_temp < img->y_size)
		f_temp +=
		  filter[(f_y * f_x_size) +
			 f_x] * img->data_list[0][(y_temp * img->x_size) +
						  x_temp];
	    }
	temp_data[(y * img->x_size) + x] = f_temp / f_div;
      }

  for (x = 0; x < img->x_size * img->y_size; x++)
    img->data_list[0][x] = temp_data[x];

  free (temp_data);
  return 0;
}




int
IMG_filter_list (IMG_list * img_list, int filter_mode)
{
  int img_nr;

  for (img_nr = 0; img_nr < img_list->total_nr; img_nr++)
    IMG_filter_img (&(img_list->img[img_nr]), filter_mode);

  fprintf (stdout, "$IMG_filter_list(...,");
  switch (filter_mode)
    {
    case IMG_sobelh:
      fprintf (stdout, "IMG_SobelH");
      break;
    case IMG_sobelv:
      fprintf (stdout, "IMG_SobelV");
      break;
    }

  fprintf (stdout, ");\n");
  return (0);
}

int
IMG_calculate_filter (double **data, int img_no, int dimension,
		      double ***filterv_data, double ***filterh_data)
{
  static const double filter[3][5] =
    { {0.0, -1.0, 0.0, 1.0, 0.0}, {-0.5, -2.0, 0.0, 2.0, 0.5}, {0.0, -1.0,
								0.0, 1.0,
								0.0}
  };

  int img_pos;
  int x, y, i, j;
  int sqrt_dim = (int) sqrt ((double) dimension);
  double sobel;

  //malloc memory for sobel_data
  *filterv_data = malloc (img_no * sizeof (double *));
  *filterh_data = malloc (img_no * sizeof (double *));
  if (*filterv_data == NULL || *filterh_data == NULL)
    {
      printf ("Nicht genuegend Speicher fuer Sobeldaten!\n");
      exit (0);
    }
  for (img_pos = 0; img_pos < img_no; img_pos++)
    {
      (*filterv_data)[img_pos] = malloc (dimension * sizeof (double));
      (*filterh_data)[img_pos] = malloc (dimension * sizeof (double));
      if ((*filterv_data)[img_pos] == NULL
	  || (*filterh_data)[img_pos] == NULL)
	{
	  printf ("Nicht genuegend Speicher fuer Sobeldaten!\n");
	  exit (0);
	}
    }

  //calculate sobel_data for each image
  for (img_pos = 0; img_pos < img_no; img_pos++)
    {
      for (y = 0; y < sqrt_dim; y++)
	for (x = 0; x < sqrt_dim; x++)
	  {
	    sobel = 0;
	    for (j = y - 1; j <= y + 1; j++)
	      for (i = x - 2; i <= x + 2; i++)
		{
		  if (j >= 0 && j < sqrt_dim && i >= 0 && i < sqrt_dim)
		    {
		      sobel +=
			data[img_pos][(sqrt_dim * j) + i] * filter[j - y +
								   1][i - x +
								      2];
		      //printf("x:%d y:%d i:%d j:%d f:%f\n",i,j,i-x+2,j-y+1,filter[j-y+1][i-x+2]);
		    }
		}
	    //printf("\n");
	    (*filterv_data)[img_pos][(sqrt_dim * y) + x] = sobel;

	    sobel = 0;
	    for (j = y - 2; j <= y + 2; j++)
	      for (i = x - 1; i <= x + 1; i++)
		{
		  if (j >= 0 && j < sqrt_dim && i >= 0 && i < sqrt_dim)
		    {
		      sobel +=
			data[img_pos][(sqrt_dim * j) + i] * filter[i - x +
								   1][j - y +
								      2];
		      //printf("-x:%d y:%d i:%d j:%d f:%f\n",i,j,i-x+1,j-y+2,filter[i-x+1][j-y+2]);
		    }
		}
	    // printf("\n");
	    (*filterh_data)[img_pos][(sqrt_dim * y) + x] = sobel;
	  }
    }
  return (0);
}

/*
double
IMG_calculate_variance (double **data, int *klasse, int img_no, int dimension)
{

  int img_pos;
  int class_pos;
  int letzte_klasse = 0;
  int i;
  double **mittel_bilder;
  double *class_var;
  double sum_var;
  int *class_anz;

  for (img_pos = 0; img_pos < img_no; img_pos++)
    {
      if (klasse[img_pos] > letzte_klasse)
	letzte_klasse = klasse[img_pos];
    }

  mittel_bilder = malloc ((letzte_klasse + 1) * sizeof (double *));
  for (class_pos = 0; class_pos <= letzte_klasse; class_pos++)
    {
      mittel_bilder[class_pos] = malloc (dimension * sizeof (double));
      for (i = 0; i < dimension; i++)
	mittel_bilder[class_pos][i] = 0.0;
    }

  class_var = malloc ((letzte_klasse + 1) * sizeof (double));
  class_anz = malloc ((letzte_klasse + 1) * sizeof (int));
  for (class_pos = 0; class_pos <= letzte_klasse; class_pos++)
    {
      class_var[class_pos] = 0.0;
      class_anz[class_pos] = 0;
    }
  // ****************************************************************
  for (img_pos = 0; img_pos < img_no; img_pos++)
    {
      for (i = 0; i < dimension; i++)
	mittel_bilder[klasse[img_pos]][i] += data[img_pos][i];
      class_anz[klasse[img_pos]]++;
    }
  for (class_pos = 0; class_pos <= letzte_klasse; class_pos++)
    {
      for (i = 0; i < dimension; i++)
	{
	  if (class_anz[class_pos] > 0)
	    mittel_bilder[class_pos][i] /= (double) class_anz[class_pos];
	}
    }
  // ******************************************************************

  for (img_pos = 0; img_pos < img_no; img_pos++)
    class_var[klasse[img_pos]] +=
      W2Deuclid_distance (data[img_pos], mittel_bilder[klasse[img_pos]],
			  dimension);

  sum_var = 0;
  printf ("\n");
  for (class_pos = 0; class_pos <= letzte_klasse; class_pos++)
    {
      class_var[class_pos] /= class_anz[class_pos];
      printf ("klasse:%d Var:%f\n", class_pos, class_var[class_pos]);
      sum_var += class_var[class_pos];
    }
  printf ("-----------------------\nSum: %f\n", sum_var);

  // *********************************************************************
  for (class_pos = 0; class_pos <= letzte_klasse; class_pos++)
    free (mittel_bilder[class_pos]);
  free (mittel_bilder);
  free (class_anz);
  free (class_var);
  // ********************************************************************
  return (sum_var);
}
*/
int*
IMG_read_number_array (char *filename)
{
  FILE *file;
  int* image_nr_array;
  int image_nr_array_length;
  int char_temp;
  int array_pos;
  int image_nr;

  if ((file = fopen (filename, "rb")) == NULL)
    {
      fprintf(stderr,"can't open file %s\n",filename);
      exit (0);
    }
  image_nr_array_length=50;
  image_nr_array=(int*) malloc((image_nr_array_length+1)*sizeof(int));
  if(image_nr_array==NULL)
    {
      fprintf (stderr, "Not enough memory for image_nr_array\n");
      exit (-1);
    }
  image_nr_array[image_nr_array_length]=-2;
  char_temp=' ';
  array_pos=0;
  while(char_temp!=EOF)
    {
      if(fscanf(file,"%d",&image_nr)==1)
	{
	  image_nr_array[array_pos]=image_nr;
	  array_pos++;
	  if(array_pos>=image_nr_array_length)
	    {
	      image_nr_array_length*=2;
	      image_nr_array=(int*) realloc(image_nr_array,(image_nr_array_length+1)*sizeof(int));
	      if(image_nr_array==NULL)
		{
		  fprintf (stderr, "Not enough memory for image_nr_array\n");
		  exit (-1);
		}
	      image_nr_array[image_nr_array_length]=-2;
	    }
	}
      else
	while('\n'!=(char_temp=getc(file)) && char_temp!=EOF);
    }
  image_nr_array[array_pos]=-1;
  fclose (file);
  return (image_nr_array);
}

int IMG_integer_compare(const void *a,const void *b)
{
  return( *((int*)a) - *((int*)b) );
}

int IMG_image_nr_array_sort(int* image_nr_array)
{
  int image_nr_array_length;

  for(image_nr_array_length=0;image_nr_array[image_nr_array_length]>=0;image_nr_array_length++);
  
  qsort(image_nr_array,image_nr_array_length,sizeof(int),IMG_integer_compare);
  return(1);
}

int
IMG_image_nr_array_print(int *image_nr_array)
{
  int array_pos;

  for(array_pos=0;image_nr_array[array_pos]>=0;array_pos++)
    {
      fprintf(stdout,"%d",image_nr_array[array_pos]);
      if(image_nr_array[array_pos+1]>=0)
	fprintf(stdout," ");
    }
  fprintf(stdout,"\n");
  
  return(1);
}

IMG_list* IMG_img_list_from_image_nr_array(IMG_list* img_list_source,int* image_nr_array)
{
  IMG_list* img_list_destination;
  int image_nr_array_length;
  int array_pos;
  int image_nr_src;
  int image_nr_dest;

  IMG_image_nr_array_sort(image_nr_array);
  
  for(image_nr_array_length=0;image_nr_array[image_nr_array_length]>=0;image_nr_array_length++);
  
  img_list_destination = malloc (sizeof (IMG_list));
  if (img_list_destination == NULL)
    {
      fprintf (stderr, "not enough memory to create IMG_list from IMG_list %s\n",
	       img_list_source->name);
      exit (-1);
    }
  img_list_destination->img = malloc (image_nr_array_length * sizeof (IMG));
  if (img_list_destination->img == NULL
      || NULL == (img_list_destination->name=
		  (char*) malloc((strlen(img_list_source->name)+1)*sizeof(char))))
    {
      fprintf (stderr, "not enough memory to create IMG_list from IMG_list %s\n",
	       img_list_source->name);
      exit (-1);
    }

  img_list_destination->total_nr = image_nr_array_length;
  
  strcpy(img_list_destination->name, img_list_source->name);

  image_nr_src=0;
  for(array_pos=0;image_nr_array[array_pos]>=0;array_pos++)
    {
      image_nr_dest=image_nr_array[array_pos];
      
      while(image_nr_src<image_nr_dest)
	image_nr_src++;
      if(image_nr_src>=img_list_source->total_nr)
	{
	  fprintf (stderr, "IMG_list %s doesn't have image_nr %d\n",
		   img_list_source->name,image_nr_src);
	  exit (-1);
	}	  
      if (IMG_copy (img_list_source->img[image_nr_src],
		    &(img_list_destination->img[array_pos])) != 0)
	{
	  fprintf (stderr, "not enough memory to copy IMG_list %s\n",
		   img_list_source->name);
	  exit (-1);
	}    
    }
  return(img_list_destination);
}
