/******************************************************************************

Copyright 2003, 2004 Christian Gollan

This file is part of the W2D package.

The W2D package is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

The W2D package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with the W2D package; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <math.h>
#include "Xdraw.h"
#include "IMG.h"
#include "W2D.h"
#include "config.h"


#ifdef USE_GDlib
int
IMG_GD_create (IMG_GD * img_gd, int x_size, int y_size)
{
  int color;

  if (img_gd->gd_image == NULL)
    {
      img_gd->gd_image = gdImageCreate (x_size, y_size);
      if (img_gd->gd_image == NULL)
	{
	  fprintf (stderr, "IMG_GD_create failed,not enough memory\n");
	  exit (-1);
	}
      //*Allocate Colours******************
      for (color = 255; color >= 0; color--)
	gdImageColorAllocate (img_gd->gd_image, color, color, color);
    }
  else
    {
      color = gdImageColorClosest (img_gd->gd_image, 255, 255, 255);
      gdImageFilledRectangle (img_gd->gd_image, 0, 0,
			      (img_gd->gd_image)->sx, (img_gd->gd_image)->sy,
			      color);
    }
  return (0);
}

int
IMG_GD_destroy (IMG_GD * img_gd)
{
  if(img_gd->gd_image!=NULL)
    {
      gdImageDestroy (img_gd->gd_image);
      img_gd->gd_image=NULL;		
    }
  return(1);
}


int
IMG_GD_draw_image (IMG_GD * img_gd, double *data, int x_size, int y_size,
		   int x_pos, int y_pos)
{
  int x, y, color;

//*start of drawing image*******************************
  for (y = 0; y < y_size; y++)
    for (x = 0; x < x_size; x++)
      {
	color = (data[(y * x_size) + x] - img_gd->min_color)
	  / ((img_gd->max_color - img_gd->min_color) / 255);
	color = gdImageColorClosest (img_gd->gd_image, color, color, color);
	gdImageFilledRectangle (img_gd->gd_image, (x * 5) + x_pos,
				(y * 5) + y_pos,
				(x * 5) + 4 + x_pos, (y * 5) + 4 + y_pos,
				color);
      }

  return (0);
}

int
IMG_GD_draw_normal_image (IMG_GD * img_gd, double *data, int x_size, int y_size,
		   int x_pos, int y_pos)
{
  int x, y, color;

//*start of drawing image*******************************
  for (y = 0; y < y_size; y++)
    for (x = 0; x < x_size; x++)
      {
	color = (data[(y * x_size) + x] - img_gd->min_color)
	  / ((img_gd->max_color - img_gd->min_color) / 255);
	color = gdImageColorClosest (img_gd->gd_image, color, color, color);
	gdImageSetPixel(img_gd->gd_image, x + x_pos, y + y_pos, color);
      }

  return (0);
}

int
IMG_GD_draw_normal_warpedimage (IMG_GD * img_gd, double *data,int data_x_size, int *warping_info,
				int x_size, int y_size, int x_pos, int y_pos)
{
  int x, y, x_temp, y_temp, color;

  if(warping_info == NULL)
    return(-1);  

//*start of drawing image*******************************
  for (y = 0; y < y_size; y++)
    for (x = 0; x < x_size; x++)
      {
	x_temp = warping_info[(((y * x_size) + x) * 2)];
	y_temp = warping_info[(((y * x_size) + x) * 2) + 1];
	color = (data[(y_temp * data_x_size) + x_temp] - img_gd->min_color)
	  / ((img_gd->max_color - img_gd->min_color) / 255);
	color = gdImageColorClosest (img_gd->gd_image, color, color, color);
	gdImageSetPixel(img_gd->gd_image, x + x_pos, y + y_pos, color);
      }

  return (0);
}

int
IMG_GD_draw_warpedimage (IMG_GD * img_gd, double *data,int data_x_size, int *warping_info,
			 int x_size, int y_size, int x_pos, int y_pos)
{
  int x, y, x_temp, y_temp, color;

  if(warping_info == NULL)
    return(-1);  

//*start of drawing image*******************************
  for (y = 0; y < y_size; y++)
    for (x = 0; x < x_size; x++)
      {
	x_temp = warping_info[(((y * x_size) + x) * 2)];
	y_temp = warping_info[(((y * x_size) + x) * 2) + 1];
	color = (data[(y_temp * data_x_size) + x_temp] - img_gd->min_color)
	  / ((img_gd->max_color - img_gd->min_color) / 255);
	color = gdImageColorClosest (img_gd->gd_image, color, color, color);
	gdImageFilledRectangle (img_gd->gd_image, (x * 5) + x_pos,
				(y * 5) + y_pos,
				(x * 5) + 4 + x_pos, (y * 5) + 4 + y_pos,
				color);
      }

  return (0);
}

int
IMG_GD_draw_warpedgrid (IMG_GD * img_gd, int *warping_info, int x_size,
			int y_size, int x_pos, int y_pos)
{
  int x, y, x1_temp, y1_temp, x2_temp, y2_temp, white, black;

  if(warping_info == NULL)
    return(-1);

  white = gdImageColorClosest (img_gd->gd_image, 255, 255, 255);
  black = gdImageColorClosest (img_gd->gd_image, 0, 0, 0);

//*start of drawing grid*******************************
  for (y = 0; y < y_size; y++)
    for (x = 0; x < x_size - 1; x++)
      {
	x1_temp = warping_info[(((y * x_size) + x) * 2)];
	y1_temp = warping_info[(((y * x_size) + x) * 2) + 1];
	x2_temp = warping_info[(((y * x_size) + x + 1) * 2)];
	y2_temp = warping_info[(((y * x_size) + x + 1) * 2) + 1];
	gdImageLine (img_gd->gd_image,
		     x_pos + 2 + (x1_temp * 5) + (x - x1_temp),
		     y_pos + 2 + (y1_temp * 5) + (y - y1_temp),
		     x_pos + 2 + (x2_temp * 5) + (x + 1 - x2_temp),
		     y_pos + 2 + (y2_temp * 5) + (y - y2_temp), black);
      }

  for (x = 0; x < x_size; x++)
    for (y = 0; y < y_size - 1; y++)
      {
	x1_temp = warping_info[(((y * x_size) + x) * 2)];
	y1_temp = warping_info[(((y * x_size) + x) * 2) + 1];
	x2_temp = warping_info[((((y + 1) * x_size) + x) * 2)];
	y2_temp = warping_info[((((y + 1) * x_size) + x) * 2) + 1];
	gdImageLine (img_gd->gd_image,
		     x_pos + 2 + (x1_temp * 5) + (x - x1_temp),
		     y_pos + 2 + (y1_temp * 5) + (y - y1_temp),
		     x_pos + 2 + (x2_temp * 5) + (x - x2_temp),
		     y_pos + 2 + (y2_temp * 5) + (y + 1 - y2_temp), black);
      }

  return (0);
}

int
IMG_GD_reset_min_max_color (IMG_GD * img_gd)
{
  //*get min- and max-color of image
  img_gd->min_color = DBL_MAX;
  img_gd->max_color = DBL_MIN;
  return (0);
}

int
IMG_GD_get_min_max_color (IMG_GD * img_gd, double *data, int dimension)
{
  int x;
  //*get min- and max-color of data
  for (x = 0; x < dimension; x++)
    {
      if (img_gd->max_color < data[x])
	img_gd->max_color = data[x];
      if (img_gd->min_color > data[x])
	img_gd->min_color = data[x];
    }
  return (0);
}

int
IMG_GD_save (IMG_GD * img_gd, char *filename)
{
  FILE *filep;

  filep = fopen (filename, "wb");
  if (filep == NULL)
    {
      fprintf (stderr, "IMG_GD_save can't create file %s\n", filename);
      exit (-1);
    }

  gdImagePng (img_gd->gd_image, filep);

  fclose (filep);

  return (0);
}

int
IMG_GD_print (IMG_GD * img_gd, int x_pos, int y_pos, char *string)
{
  int black;

  black = gdImageColorClosest (img_gd->gd_image, 0, 0, 0);
  gdImageString (img_gd->gd_image, gdFontLarge, x_pos, y_pos, string, black);
  return (0);
}

/*
int
IMG_GD_W2D_draw (IMG_GD * img_gd, double *dataA, double *dataB,
		 int *warping_info, int x_size, int y_size, char *string)
{
  IMG_GD_create (img_gd, x_size * 5 * 4 + 1, y_size * 5 + 20);
  IMG_GD_reset_min_max_color (img_gd);
  IMG_GD_get_min_max_color (img_gd, dataA, x_size * y_size);
  IMG_GD_get_min_max_color (img_gd, dataB, x_size * y_size);
  IMG_GD_draw_image (img_gd, dataA, x_size, y_size, 0, 0);
  IMG_GD_draw_warpedimage (img_gd, dataB, warping_info, x_size, y_size,
			   x_size * 5 + 1, 0);
  IMG_GD_draw_warpedgrid (img_gd, warping_info, x_size, y_size,
			  x_size * 5 * 2 + 1, 0);
  IMG_GD_draw_image (img_gd, dataB, x_size, y_size, x_size * 5 * 3 + 1, 0);
  IMG_GD_print (img_gd, 0, y_size * 5, string);
  return (0);
}
*/
int
IMG_GD_draw_warping_images (IMG_GD * img_gd, IMG A, IMG B, int *warping_info,
			    char *string)
{
  int vis_pos;
  int dis_pos;
  int max_x_size;
  int max_y_size;

  max_x_size=A.x_size;
  max_y_size=A.y_size;

  if(max_x_size<B.x_size)
	max_x_size=B.x_size;
  if(max_y_size<B.y_size)
	max_y_size=B.y_size;

  vis_pos = 0;
  dis_pos = 0;
  while (B.dis_mode[dis_pos] >= 0)
    {
      if (B.data_length[dis_pos] == B.x_size*B.y_size)
	vis_pos++;
      dis_pos++;
    }
  if(img_gd->gd_image!=NULL)
	{
        gdImageDestroy (img_gd->gd_image);
	img_gd->gd_image=NULL;		
	}
  IMG_GD_create (img_gd, max_x_size * 5 * 4 + 1,
		 (((max_y_size * 5) + 1) * vis_pos) + 20);
  dis_pos = 0;
  vis_pos = 0;
  while (B.dis_mode[dis_pos] >= 0)
    {
      if (B.data_length[dis_pos] == B.x_size*B.y_size)
	{
	  IMG_GD_reset_min_max_color (img_gd);
	  IMG_GD_get_min_max_color (img_gd, A.data_list[dis_pos],
				    A.data_length[dis_pos]);
	  IMG_GD_get_min_max_color (img_gd, B.data_list[dis_pos],
				    B.data_length[dis_pos]);
	  IMG_GD_draw_image (img_gd, A.data_list[dis_pos], A.x_size, A.y_size,
			     0, ((max_y_size * 5) + 1) * vis_pos);
	  IMG_GD_draw_warpedimage (img_gd, B.data_list[dis_pos],B.x_size, warping_info,
				   A.x_size, A.y_size, max_x_size * 5 + 1,
				   ((max_y_size * 5) + 1) * vis_pos);
	  IMG_GD_draw_warpedgrid (img_gd, warping_info, A.x_size, A.y_size,
				  max_x_size * 5 * 2 + 1,
				  ((max_y_size * 5) + 1) * vis_pos);
	  IMG_GD_draw_image (img_gd, B.data_list[dis_pos], B.x_size, B.y_size,
			     max_x_size * 5 * 3 + 1,
			     ((max_y_size * 5) + 1) * vis_pos);
	  vis_pos++;
	}
      dis_pos++;
    }
  IMG_GD_print (img_gd, 0, ((max_y_size * 5) + 1) * vis_pos, string);
  return (0);
}

int
IMG_GD_draw_img (IMG_GD * img_gd, IMG A)
{
  int vis_pos;
  int dis_pos;

  vis_pos = 0;
  dis_pos = 0;
  while (A.dis_mode[dis_pos] > -2)
    {
      if (A.data_length[dis_pos] == (A.x_size*A.y_size))
	vis_pos++;
      dis_pos++;
    }
  if(img_gd->gd_image!=NULL)
    {
      gdImageDestroy (img_gd->gd_image);
      img_gd->gd_image=NULL;		
    }
  IMG_GD_create (img_gd, A.x_size * 5,
		 (((A.y_size * 5) + 1) * vis_pos) - 1);
  dis_pos = 0;
  vis_pos = 0;
  while (A.dis_mode[dis_pos] > -2)
    {
      if (A.data_length[dis_pos] == (A.x_size*A.y_size))
	{
	  IMG_GD_reset_min_max_color (img_gd);
	  IMG_GD_get_min_max_color (img_gd, A.data_list[dis_pos],
				    A.data_length[dis_pos]);
	  IMG_GD_draw_image (img_gd, A.data_list[dis_pos], A.x_size, A.y_size,
			     0, ((A.y_size * 5) + 1) * vis_pos);
	  vis_pos++;
	}
      dis_pos++;
    }
  return (0);
}


int IMG_GD_min_max_color_img_list(IMG_GD* img_gd,IMG_list* img_list)
{
  int img_nr;
  int img_depth;

  IMG_GD_reset_min_max_color (img_gd);
  for(img_nr=0;img_nr<img_list->total_nr;img_nr++)
    for(img_depth=0;
	img_depth<=IMG_get_depth(&img_list->img[img_nr]);img_depth++)
      if(img_list->img[img_nr].data_length[img_depth]
	 ==img_list->img[img_nr].x_size*img_list->img[img_nr].y_size)      
	IMG_GD_get_min_max_color (img_gd,
				  img_list->img[img_nr].data_list[img_depth],
				  img_list->img[img_nr].data_length[img_depth]);
  return(0);
}


int IMG_GD_save_img_png(IMG img,char* png_file_name)
{
  int img_depth;
  int img_save_depth;
  IMG_GD img_gd;

  img_gd.gd_image=NULL;

  img_save_depth=0;
  for(img_depth=0;img_depth<=IMG_get_depth(&img);img_depth++)
    if(img.data_length[img_depth]==img.x_size*img.y_size)
      img_save_depth++;
 
  IMG_GD_create (&img_gd, img.x_size, img.y_size*img_save_depth);
  img_gd.max_color=1.0;
  img_gd.min_color=0.0;

  img_save_depth=0;
  for(img_depth=0;img_depth<=IMG_get_depth(&img);img_depth++)
    if(img.data_length[img_depth]==img.x_size*img.y_size)      
      {	
	IMG_GD_draw_normal_image (&img_gd, img.data_list[img_depth],
			   img.x_size, img.y_size,
			   0, img.y_size*img_save_depth);
	img_save_depth++;
      }
  IMG_GD_save (&img_gd, png_file_name);

  gdImageDestroy(img_gd.gd_image);
  return (0);
}

int IMG_GD_save_warped_img_png(IMG A, IMG B, int* warping, char* png_file_name)
{
  int img_depth;
  int img_save_depth;
  IMG_GD img_gd;

  img_gd.gd_image=NULL;

  img_save_depth=0;
  for(img_depth=0;img_depth<=IMG_get_depth(&B);img_depth++)
    if(B.data_length[img_depth]==B.x_size*B.y_size)
      img_save_depth++;
 
  IMG_GD_create (&img_gd, A.x_size, A.y_size*img_save_depth);
  img_gd.max_color=1.0;
  img_gd.min_color=0.0;

  img_save_depth=0;
  for(img_depth=0;img_depth<=IMG_get_depth(&B);img_depth++)
    if(B.data_length[img_depth]==B.x_size*B.y_size)      
      {	
		IMG_GD_draw_normal_warpedimage (&img_gd, B.data_list[img_depth], B.x_size, warping,
					A.x_size, A.y_size,
					0, A.y_size*img_save_depth);
	img_save_depth++;
      }
  IMG_GD_save (&img_gd, png_file_name);

  gdImageDestroy(img_gd.gd_image);
  return (0);
}

#ifdef USE_X11lib
int
IMG_X_create (IMG_GD * img_gd)
{
  int i;
  XColor color;
  Colormap colormap;

  img_gd->dis = XOpenDisplay (NULL);
  if(img_gd->dis==NULL)
    {
      fprintf(stderr,"Failed to open display\n");
      exit(-1);
    }
  img_gd->win =
    XCreateSimpleWindow (img_gd->dis, RootWindow (img_gd->dis, 0), 0, 0,
			 img_gd->gd_image->sx, img_gd->gd_image->sy, 0,
			 WhitePixel (img_gd->dis, 0), WhitePixel (img_gd->dis,
								  0));

  XMapWindow (img_gd->dis, img_gd->win);
  colormap = DefaultColormap (img_gd->dis, 0);
  img_gd->my_gc = XCreateGC (img_gd->dis, img_gd->win, 0, 0);
  for (i = 0; i < 256; i++)
    {
      color.red = i * 256;
      color.green = i * 256;
      color.blue = i * 256;
      XAllocColor (img_gd->dis, colormap, &color);
      img_gd->colormap[255 - i] = color.pixel;
    }
  XSelectInput (img_gd->dis, img_gd->win,
		ExposureMask | KeyPressMask | ButtonPressMask);

  XFlush (img_gd->dis);
  return (0);
}

int
IMG_X_draw (IMG_GD * img_gd)
{
  int x, y;
  int x_size, y_size;
  int color;
  gdImagePtr temp_gd_image;
  Window root_return;
  int x_return;
  int y_return;
  unsigned int width_return;
  unsigned int height_return;
  unsigned int border_width_return;
  unsigned int depth_return;

  x_size = img_gd->gd_image->sx;
  y_size = img_gd->gd_image->sy;

  if (img_gd->dis == NULL)
    IMG_X_create (img_gd);

  XGetGeometry (img_gd->dis, img_gd->win, &root_return, &x_return, &y_return,
		&width_return, &height_return, &border_width_return,
		&depth_return);

  if (x_size != (int) width_return || y_size != (int) height_return)
    {
      temp_gd_image = gdImageCreate (width_return, height_return);
      for (color = 255; color >= 0; color--)
	gdImageColorAllocate (temp_gd_image, color, color, color);
      gdImageCopyResized (temp_gd_image, img_gd->gd_image, 0, 0, 0, 0,
			  temp_gd_image->sx, temp_gd_image->sy,
			  img_gd->gd_image->sx, img_gd->gd_image->sy);
      for (y = 0; y < temp_gd_image->sy; y++)
	for (x = 0; x < temp_gd_image->sx; x++)
	  {
	    color = gdImageGetPixel (temp_gd_image, x, y);
	    XSetForeground (img_gd->dis, img_gd->my_gc,
			    img_gd->colormap[color]);
	    XDrawPoint (img_gd->dis, img_gd->win, img_gd->my_gc, x, y);
	  }
      gdImageDestroy (temp_gd_image);
    }
  else
    {
      for (y = 0; y < y_size; y++)
	for (x = 0; x < x_size; x++)
	  {
	    color = gdImageGetPixel (img_gd->gd_image, x, y);
	    XSetForeground (img_gd->dis, img_gd->my_gc,
			    img_gd->colormap[color]);
	    XDrawPoint (img_gd->dis, img_gd->win, img_gd->my_gc, x, y);
	  }
    }
  XFlush (img_gd->dis);
  //  XSetForeground(img_gd->dis, img_gd->my_gc, img_gd->colormap[0]);
  //  XClearWindow(img_gd->dis, img_gd->win);

  return (0);
}
#endif //USE_X11lib
#endif //USE_GDlib
