/******************************************************************************

Copyright 2003, 2004 Christian Gollan

This file is part of the W2D package.

The W2D package is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

The W2D package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with the W2D package; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

******************************************************************************/

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <time.h>
#include <argp.h>
#include "IMG.h"
#include "W2D.h"
#include "W2DW.h"
#include "W2DHM.h"
#include "W2DSA.h"
#include "W2DHUNG.h"
#include "W2DTD.h"
#include "Xdraw.h"
#include "DIS.h"


#define STRING_BUFFER_LENGTH 400 //has to be at least 50 for agrp distance parsing
#define DIS_MODE_LIST_LENGTH 100 //define max. length for Dis_mode_list (image:euc2:1,penalty:pen:1,...)

int
get_keyinteger (char *keyword, char *type)
{
  int x = 0;
  while (keylist[x].string != NULL)
    {
      if (!strcmp (keylist[x].string, keyword)
	  && !strcmp (keylist[x].type, type))
	return (keylist[x].integer);
      x++;
    }
  return (-1);
}

char *
get_keystring (int integer, char *type)
{
  int x = 0;
  while (keylist[x].string != NULL)
    {
      if (keylist[x].integer == integer && !strcmp (keylist[x].type, type))
	return (keylist[x].string);
      x++;
    }
  return (NULL);
}

const char *argp_program_version =
VERSION;

const char *argp_program_bug_address =
PACKAGE_BUGREPORT;

/* This structure is used by main to communicate with parse_opt. */
struct arguments
{
  char *args[2];            /* ARG1 */
  int test_img_counter;
  int train_img_counter;
  int test_img_start;
  int train_img_start;
  int NN_counter;
  int NN_mode;
  int beamsize;
  double dis_limit;
  int warprange;
  int Xdraw;
  int write_gif;
  int leaving_one_out;
  int bordersize;
  int flip_mode;
  int shift;
  int W2D_alg;
  double dis_mode_list[DIS_MODE_LIST_LENGTH];
  int dis_mode_list_counter;
  int SYM_dis;
  int cluster_mode;
  char* save_train_data_name;
  char* disdat_file_name;
  int norm_mode;
  int wait_mode;
  int test_class;
  int train_class;
  int dis5x5_var;
  int save_warped_image;
};

void init_arguments(struct arguments *arguments)
{
  arguments->test_img_counter = -1;
  arguments->train_img_counter = 0;
  arguments->test_img_start = 0;
  arguments->train_img_start = 0;
  arguments->NN_counter = -1;
  arguments->NN_mode = 0;
  arguments->beamsize = -1;
  arguments->dis_limit = -1;
  arguments->warprange = -1;
  arguments->Xdraw = 0;
  arguments->write_gif = 0;
  arguments->leaving_one_out = 0;
  arguments->bordersize = 0;
  arguments->flip_mode = 0;
  arguments->shift = 1;
  arguments->W2D_alg = W2D_NO;
  arguments->dis_mode_list_counter = 0;
  arguments->SYM_dis = 0;
  arguments->cluster_mode = -1;
  arguments->save_train_data_name=NULL;
  arguments->norm_mode=0;
  arguments->wait_mode=0;
  arguments->disdat_file_name=NULL;
  arguments->test_class=-1;
  arguments->train_class=-1;
  arguments->dis5x5_var=0;
  arguments->save_warped_image=0;
}


/*
   OPTIONS.  Field 1 in ARGP.
   Order of fields: {NAME, KEY, ARG, FLAGS, DOC}.
*/
static struct argp_option options[] =
{
  {"trainstart", 'A', "N", 0, "First train image from [FILE1]"}, 
  {"traintotal",'R', "N", 0, "Number of train images to load"},
  {"trainclass",'H', "N", 0, "Class for train images"},
  {"teststart",'S',"N",0, "First test image from [FILE2]"},
  {"testtotal",'E',"N",0, "Number of test images to load"},
  {"testclass",'G',"N",0, "Class for test images"},  
  {"algorithm",'M',"(WW2DHM/P2DHM/P2DHMD/SA2DHM/ID/ID3/HUNG/TD/NO)",0, "Warpingmodel (default: NO)"},
  {"nearest",'N',"N",0, "Nearest neighbour list size; uses the features as given by -D but does no warping; use a precomputed list with -L if you want e.g. plain Euclidean distance of grayvalues"},
  {"nearestmode",'V',"N",0, "Nearest neighbour mode"},
  {"data",'D',"data:mode:scale",0,"with data in: file, image, nullimg, sobelv, sobelh, sobeld, laplace, penalty, posx, posy;  mode in: abs, abs3x3, euc2, euc3, euc2rec3x3, euc3rec3x3, euc2rec5x5, euc2bin3x3, euc2bin5x5, euc2rec7x7, euc2rec9x9, euc2rec11x11, tan2rec3x3, euc2thresholdrec3x3:threshold, warpsum:0.0, pen, pen2, len, len2, penborder, penwarp, pixelthreshold, pixeldismove; scale: number"},
  {"warprange",'W',"N",0, "Warprange for a pixelcoordinate (default for IDM is 1, otherwise INF)"},
  {"beamsize",'B',"N",0, "Beamsize for distance algorithm"},
  {"threshold",'K',"N.N",0, "Threshold for distance algorithm"},
  {"addborder",'O',"N",0, "Add border to the images"},
  {"clustermode",'C',"N",0, "Clustermode"},
  {"trainsave",'F',"[FILE]",0, "Save modified train data"},
  {"disdat",'L',"[FILE]",0, "Read disdatfile"}, 
  {"normmode",'Q',"N",0, "Normmode(0=all images/1=seperate train and test/2=no)"},
  {"shift",'s',0,0, "Shift 3x3 train data"},
  {"vflip",'v',0,0, "Flip all images vertically"},  
  {"hflip",'h',0,0, "Flip all images horizontally"},
  {"rotate",'r',0,0, "Rotate all images 90 degrees clockwise"},
  {"symdis",'y',0,0, "Symmetric distance"},
  {"pause",'p',0,0, "Wait after each image"},
  {"write",'w',0,0, "Write PNG files from images and result"},
  {"show",'x',0,0, "Show images on X11"},  
  {"var5x5",'a',0,0, "Pixeldistance 5x5 variance"},  
  {"result",'d',0,0, "Write PNG files from result"},  
  {0}
};

/*
   PARSER. Field 2 in ARGP.
   Order of parameters: KEY, ARG, STATE.
*/
static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
  struct arguments *arguments = state->input;
  char string_buffer[STRING_BUFFER_LENGTH];
  float tmp_float;

  switch (key)
    {
    case 'M':
      if ((arguments->W2D_alg = get_keyinteger (arg, "W2D_alg")) == -1)
	{
	  argp_usage(state);
	  return -1;
	}
      break;
    case 'O':
      if (sscanf (arg, "%d%s", &(arguments->bordersize), arg) != 1)
	{
	  argp_usage(state);
	  return -1;
	}
      break;
    case 'R':
      if (sscanf (arg, "%d%s", &(arguments->train_img_counter), arg) != 1
	  || arguments->train_img_counter<=0)
	{
	  argp_usage(state);
	  return -1;
	}
      break;
    case 'E':
      if (sscanf (arg, "%d%s", &(arguments->test_img_counter), arg) != 1
	  || arguments->test_img_counter<0)
	{
	  argp_usage(state);
	  return -1;
	}
      break;
    case 'G':
      if (sscanf (arg, "%d%s", &(arguments->test_class), arg) != 1)
	{
	  argp_usage(state);
	  return -1;
	}
      break;
    case 'H':
      if (sscanf (arg, "%d%s", &(arguments->train_class), arg) != 1)
	{
	  argp_usage(state);
	  return -1;
	}
      break;
    case 'D':
      while (1)
	{
	  if (sscanf (arg, "%50[^:]:%s", string_buffer, arg) < 2)
	    {
	      argp_usage(state);
	      return -1;
	    }
	  arguments->dis_mode_list[(arguments->dis_mode_list_counter) * 3] =
	    (double) get_keyinteger (string_buffer, "IMG_mode");
	  if (arguments->dis_mode_list[(arguments->dis_mode_list_counter) * 3] == -1)
	    {
	      argp_usage(state);
	      return -1;
	    }
	  
	  if (sscanf (arg, "%50[^:]:%f%s", string_buffer, &tmp_float, arg) <
	      2)
	    {
	      argp_usage(state);
	      return -1;
	    }
	  
	  arguments->dis_mode_list[(arguments->dis_mode_list_counter * 3) + 1] =
	    (double) get_keyinteger (string_buffer, "W2D_dis_mode");
	  if (arguments->dis_mode_list[(arguments->dis_mode_list_counter * 3) + 1] == -1
	      || (arguments->dis_mode_list[(arguments->dis_mode_list_counter) * 3] == IMG_penalty
		  && (arguments->dis_mode_list[(arguments->dis_mode_list_counter * 3) + 1] >= 1000
		      || arguments->dis_mode_list[(arguments->dis_mode_list_counter * 3) + 1] < 500))
	      || (arguments->dis_mode_list[(arguments->dis_mode_list_counter) * 3] == IMG_image
		  && (arguments->dis_mode_list[(arguments->dis_mode_list_counter * 3) + 1] >= 2000
		      || arguments->dis_mode_list[(arguments->dis_mode_list_counter * 3) + 1] < 1000)))
	    {
	      argp_usage(state);
	      return -1;
	    }
	  arguments->dis_mode_list[(arguments->dis_mode_list_counter * 3) + 2] = tmp_float;
	  
	  arguments->dis_mode_list_counter++;
	  if(arguments->dis_mode_list_counter>=DIS_MODE_LIST_LENGTH)
	    {
	      fprintf(stderr,"The programm was compiled for a maximal Dis_mode_list length %d\n",DIS_MODE_LIST_LENGTH);
	      return -1;
	    }
	  if (!sscanf (arg, ",%s", arg))
	    break;
	}
      break;
    case 'B':
      if (sscanf (arg, "%d%s", &(arguments->beamsize), arg) != 1)
	{
	  argp_usage(state);
	  return -1;
	}
      break;
    case 'K':
      if (sscanf (arg, "%f%s", &tmp_float, arg) != 1)
	{
	  argp_usage(state);
	  return -1;
	}
      arguments->dis_limit= (double) tmp_float;
      break;
    case 'W':
      if (sscanf (arg, "%d%s", &(arguments->warprange), arg) != 1)
	{
	  argp_usage(state);
	  return -1;
	}
      break;
    case 'N':
      if (sscanf (arg, "%d%s", &(arguments->NN_counter), arg) != 1)
	{
	  argp_usage(state);
	  return -1;
	}
      break;
    case 'V':
      if (sscanf (arg, "%d%s", &(arguments->NN_mode), arg) != 1
	  || arguments->NN_mode<0)
	{
	  argp_usage(state);
	  return -1;
	}
      break;
    case 'S':
      if (sscanf (arg, "%d%s", &(arguments->test_img_start), arg) != 1)
	{
	  argp_usage(state);
	  return -1;
	}
      break;
    case 'A':
      if (sscanf (arg, "%d%s", &(arguments->train_img_start), arg) != 1)
	{
	  argp_usage(state);
	  return -1;
	}
      break;
    case 'C':
      if (sscanf (arg, "%d%s", &(arguments->cluster_mode), arg) != 1
	  || arguments->cluster_mode<0)
	    {
	      argp_usage(state);
	      return -1;
	    }
	  break;
    case 'F':
      arguments->save_train_data_name=malloc((strlen(arg)+1)*sizeof(char));
      if (arguments->save_train_data_name==NULL || strlen(arg)<=0)
	{
	  argp_usage(state);
	  return -1;
	}
      strcpy(arguments->save_train_data_name,arg);
      break;
    case 'L':
      if (strlen(arg)<=0)
	{
	  argp_usage(state);
	  return -1;
	}
      arguments->disdat_file_name=arg;
      break;
    case 'Q':
      if (sscanf (arg, "%d%s", &(arguments->norm_mode), arg) != 1
	  || arguments->norm_mode<-1)
	{
	  argp_usage(state);
	  return -1;
	}
      break;	  
    case 'w':
      arguments->write_gif = 1;
      break;
    case 'x':
      arguments->Xdraw = 1;
      break;
    case 'v':
      arguments->flip_mode = arguments->flip_mode | IMG_VERTICAL;
      break;
    case 'h':
      arguments->flip_mode = arguments->flip_mode | IMG_HORIZONTAL;
      break;
    case 'r':
      arguments->flip_mode = arguments->flip_mode | IMG_ROTATE;
      break;
    case 's':
      arguments->shift = 9;
      break;
    case 'y':
      arguments->SYM_dis = 1;
      break;
    case 'p':
      arguments->wait_mode = 1;
      break;
    case 'a':
      arguments->dis5x5_var = 1;
      break;
    case 'd':
      arguments->save_warped_image = 1;
      break;
    case ARGP_KEY_ARG:
      if (state->arg_num >= 2)
	{
	  argp_usage(state);
	}
      arguments->args[state->arg_num] = arg;
      break;
    case ARGP_KEY_END:
      if (state->arg_num != 2)
	{
	  argp_usage (state);
	}
      else
	{
	  if(arguments->dis_mode_list_counter == 0)
	    {
	      arguments->dis_mode_list[0] =
		(double) get_keyinteger ("image", "IMG_mode");
	      arguments->dis_mode_list[1] =
		(double) get_keyinteger ("euc2", "W2D_dis_mode");
	      arguments->dis_mode_list[2] = 1.0;
	      arguments->dis_mode_list_counter=1;
	    }
	  if(  (arguments->W2D_alg == W2D_IDM) && (arguments->warprange < 0) )
	    arguments->warprange=1;
	}
      break;
    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

/*
   ARGS_DOC. Field 3 in ARGP.
   A description of the non-option command-line arguments
     that we accept.
*/
static char args_doc[] = "[FILE1] [FILE2]";

/*
  DOC.  Field 4 in ARGP.
  Program documentation.
*/
static char doc[] =
"W2D -- The program reads train FILE1 and test FILE2, the output are distances.\vExample:\n W2D train.ejf.gz test.jf -MP2DHM -Dimage:euc2:1,sobelh:euc2:0.5,sobelv:euc2:0.5,penalty:pen:1";

/*
   The ARGP structure itself.
*/
static struct argp argp = {options, parse_opt, args_doc, doc};


int
print_time_date (FILE * filep)
{
  time_t time_var;

  time (&time_var);
  return (fprintf (filep, "%s", ctime (&time_var)));
}

int
main (int argc, char *argv[])
{
  struct arguments arguments;
  double *NN_dis;
  int *NN_img;
  double NN_best_dis;
  int NN_best_class;
  double dis, dis_wop;
  int i, j, z;
  char string_buffer[STRING_BUFFER_LENGTH];
  int int_temp;
  int *warping_info;
  double min, max;
  int img_depth;

  IMG_list *img_list_tr;
  IMG_list *img_list_te;
  IMG_list *img_list_tr_tmp;
  IMG_list *img_list_te_tmp = NULL;
  IMG_list *img_list_temp = NULL;
  IMG* A;
  IMG* B;

  double (*W2D_function)(IMG,IMG,int*);

  /**********************/
  char *tr_name;
  char *te_name;
  int test_img_counter;
  int train_img_counter;
  int test_img_start;
  int train_img_start;
  int NN_counter;
  int NN_mode;
  int beamsize;
  double dis_limit;
  int warprange;
  int Xdraw;
  int write_gif;
  int leaving_one_out;
  int bordersize;
  int flip_mode;
  int shift;
  int W2D_alg;
  double* dis_mode_list;
  int dis_mode_list_counter;
  int SYM_dis;
  int cluster_mode;
  char* save_train_data_name;
  char* disdat_file_name;
  int norm_mode;
  int wait_mode;
  DISDAT disdat;
  int disdat_offset;
  int test_class;
  int train_class;
  int W2D_penvar_mode;
  int W2D_warpsum_mode;
  int dis5x5_var;
  int save_warped_image;
  /**********************/

  #ifdef USE_GDlib
  IMG_GD img_gd;
  img_gd.gd_image = NULL;
  img_gd.dis = NULL;
  #endif

  setbuf (stdout, NULL);

  snprintf (string_buffer,STRING_BUFFER_LENGTH,"#");
  for (i = 0; i < argc; i++)
    snprintf (&(string_buffer[strlen(string_buffer)]),
	      (STRING_BUFFER_LENGTH-strlen(string_buffer)),"%s ", argv[i]);

  init_arguments(&arguments);  
  argp_parse (&argp, argc, argv, 0, 0, &arguments);

  fprintf(stdout,"%s\n",string_buffer);


  tr_name=arguments.args[0];
  te_name=arguments.args[1];
  test_img_counter=arguments.test_img_counter;
  train_img_counter=arguments.train_img_counter;
  test_img_start=arguments.test_img_start;
  train_img_start=arguments.train_img_start;
  NN_counter=arguments.NN_counter;
  NN_mode=arguments.NN_mode;
  beamsize=arguments.beamsize;
  dis_limit=arguments.dis_limit;
  warprange=arguments.warprange;
  Xdraw=arguments.Xdraw;
  write_gif=arguments.write_gif;
  leaving_one_out=arguments.leaving_one_out;
  bordersize=arguments.bordersize;
  flip_mode=arguments.flip_mode;
  shift=arguments.shift;
  W2D_alg=arguments.W2D_alg;
  dis_mode_list=arguments.dis_mode_list;
  dis_mode_list_counter=arguments.dis_mode_list_counter;
  SYM_dis=arguments.SYM_dis;
  cluster_mode=arguments.cluster_mode;
  save_train_data_name=arguments.save_train_data_name;
  norm_mode=arguments.norm_mode;
  wait_mode=arguments.wait_mode;
  disdat_file_name=arguments.disdat_file_name;
  test_class=arguments.test_class;
  train_class=arguments.train_class;
  W2D_penvar_mode=0;
  W2D_warpsum_mode=0;
  dis5x5_var=arguments.dis5x5_var;
  save_warped_image=arguments.save_warped_image;

  fprintf (stdout, "#Start_time: ");
  print_time_date (stdout);
  fprintf (stdout, "#W2D_version: %s\n", VERSION);

  //***************read image files**************************
  img_list_tr = IMG_load_list (tr_name);

  if (img_list_tr == NULL)
    exit (-1);
  if (strcmp (tr_name, te_name) == 0)
    {
      fprintf (stdout, "#same files => leaving_one_out\n");
      leaving_one_out = 1;
      img_list_te = img_list_tr;
    }
  else
    {
      img_list_te = IMG_load_list (te_name);
      if (img_list_te == NULL)
	exit (-1);
    }

  //******2dw***************
  fprintf (stdout,
	   "#Traindata: %d, Testdata: %d, Trainstart: %d, Teststart: %d\n",
	   img_list_tr->total_nr, img_list_te->total_nr, train_img_start,
	   test_img_start);
  fprintf (stdout, "#W2D_mode: %s\n", get_keystring (W2D_alg, "W2D_alg"));
  if(SYM_dis==1)
    fprintf (stdout, "#Symmetric Distance\n");
  fprintf (stdout, "#NN_counter: %d\n", NN_counter);
  if (write_gif == 1)
    {
      fprintf (stdout, "#write warped images in seperate files\n");
#ifndef USE_GDlib
      fprintf(stderr,"Not compiled with GD support.\n");
#endif
    }
  if (Xdraw == 1)
    {
      fprintf (stdout, "#show warped images on X\n");
#ifndef USE_X11lib
      fprintf(stderr,"Not compiled with X11 support.\n");
#endif
    }
  if(disdat_file_name!=NULL)
    {
      disdat=DIS_read_disdat_file(disdat_file_name);
    }
  if(norm_mode==0)
    {
      for(img_depth=0;img_depth<=IMG_get_depth_list(img_list_tr);img_depth++)
	{
	  min = DBL_MAX;
	  max = -DBL_MAX;
	  IMG_get_min_max_list (img_list_tr, img_depth, &min, &max);
	  if (img_list_te != img_list_tr)
	    IMG_get_min_max_list (img_list_te, img_depth, &min, &max);
	  IMG_norm_list (img_list_tr, img_depth, min, max, 0.0, 1.0);
	  if (img_list_te != img_list_tr)
	    IMG_norm_list (img_list_te, img_depth, min, max, 0.0, 1.0);
	}
    }
  if(norm_mode==1)
    {
      for(img_depth=0;img_depth<=IMG_get_depth_list(img_list_tr);img_depth++)
	{
	  min = DBL_MAX;
	  max = -DBL_MAX;
	  IMG_get_min_max_list (img_list_tr, img_depth, &min, &max);
	  IMG_norm_list (img_list_tr, img_depth, min, max, 0.0, 1.0);
	}
      if (img_list_te != img_list_tr)
	{
	  for(img_depth=0;img_depth<=IMG_get_depth_list(img_list_te);img_depth++)
	    {
	      min = DBL_MAX;
	      max = -DBL_MAX;
	      IMG_get_min_max_list (img_list_te, img_depth, &min, &max);
	      IMG_norm_list (img_list_te, img_depth, min, max, 0.0, 1.0);	  
	    }
	}
    }

//flip_images***************************************************************************
  if (flip_mode != 0)
    {
      IMG_flip_list (img_list_tr, flip_mode);
      if (img_list_te != img_list_tr)
	IMG_flip_list (img_list_te, flip_mode);
    }
//*************************************************************************************

  //make image borders*****************************************************************8
  if (bordersize > 0)
    {
      IMG_add_border_list (img_list_tr, bordersize, bordersize, bordersize,
			   bordersize, 0.0);
      if (img_list_te != img_list_tr)
	IMG_add_border_list (img_list_te, bordersize, bordersize, bordersize,
			     bordersize, 0.0);
    }
  //***********************************************************************************8

  //shift3x3**************************************************************************
  if (shift == 9)
    {
      img_list_tr_tmp = img_list_tr;
      img_list_tr = IMG_shift3x3_list (img_list_tr, 0.0);
      if (img_list_te != img_list_tr_tmp)
	IMG_free_list (img_list_tr_tmp);
    }
  //**********************************************************************************

  //***********************************************************

  img_list_tr_tmp =
    IMG_create_empty_list (img_list_tr, dis_mode_list_counter);
  // IMG_print_list_info (img_list_tr_tmp);
  if (img_list_te != img_list_tr)
    {
      img_list_te_tmp =
	IMG_create_empty_list (img_list_te, dis_mode_list_counter);
      // IMG_print_list_info (img_list_te_tmp);
    }

  img_depth=0;
  for (i = 0; i < dis_mode_list_counter; i++)
    {
      printf ("#IMG_mode:%s",
	      get_keystring ((int) dis_mode_list[(i * 3)], "IMG_mode"));
      printf (" Dis_mode:%s %f\n",
	      get_keystring ((int) dis_mode_list[(i * 3) + 1],
			     "W2D_dis_mode"), dis_mode_list[(i * 3) + 2]);

      switch ((int) dis_mode_list[(i * 3)])
	{
	case IMG_file:
	  if(img_depth>IMG_get_depth_list(img_list_tr))
	    {
	      fprintf(stderr,"IMG_list: %s does not have the img_depth:%d\n",img_list_tr->name,img_depth);
	      exit(-1);
	    }
	  IMG_move_in_list(img_list_tr,img_depth,img_list_tr_tmp,i);
	  if(img_list_te != img_list_tr)
	    {
	      IMG_move_in_list(img_list_te,img_depth,img_list_te_tmp,i);
	       if(img_depth>IMG_get_depth_list(img_list_te))
		 {
		   fprintf(stderr,"IMG_list: %s does not have the img_depth:%d\n",img_list_te->name,img_depth);
		   exit(-1);		   
		 }
	    }
	  img_depth++;
	  break;
	case IMG_image:
	  // IMG_print_list_info (img_list_tr);
	  img_list_temp = IMG_copy_list (img_list_tr);
	  // IMG_print_list_info(img_list_temp);
	  IMG_insert_in_list (img_list_temp, img_list_tr_tmp, i);
	  IMG_free_list (img_list_temp);
	  if (img_list_te != img_list_tr)
	    {
	      img_list_temp = IMG_copy_list (img_list_te);
	      IMG_insert_in_list (img_list_temp, img_list_te_tmp, i);
	      IMG_free_list (img_list_temp);
	    }
	  break;
	case IMG_nullimg:
	  img_list_temp = IMG_copy_list (img_list_tr);
	  IMG_insert_in_list (img_list_temp, img_list_tr_tmp, i);
	  IMG_free_list (img_list_temp);
	  IMG_set_data_in_list (img_list_tr_tmp, i, -1.0);
	  if (img_list_te != img_list_tr)
	    {
	      img_list_temp = IMG_copy_list (img_list_te);
	      IMG_insert_in_list (img_list_temp, img_list_te_tmp, i);
	      IMG_free_list (img_list_temp);
	      IMG_set_data_in_list (img_list_te_tmp, i, -1.0);
	    }
	  if((int) dis_mode_list[(i * 3) + 1]==W2D_penvar)
	    W2D_penvar_mode=1;
	  if((int) dis_mode_list[(i * 3) + 1]==W2D_warpsum)
	    W2D_warpsum_mode=1;	  
	  break;
	case IMG_penalty:
	  if((int) dis_mode_list[(i * 3) + 1]==W2D_euc2rec3x3var)
	    {
	      IMG_add_dataspace_in_list(img_list_tr_tmp, i, 9);
	      IMG_set_data_in_list (img_list_tr_tmp, i, 1.0);
	      if (img_list_te != img_list_tr)
		{
		  IMG_add_dataspace_in_list(img_list_te_tmp, i, 9);
		  IMG_set_data_in_list (img_list_te_tmp, i, 1.0);
		}
	    }
	  break;

    case IMG_position_x:

      img_list_temp=IMG_copy_list(img_list_tr);
      IMG_add_position_x_list(img_list_temp);
      IMG_insert_in_list(img_list_temp, img_list_tr_tmp,i);
      IMG_free_list(img_list_temp);
	  IMG_norm_list (img_list_tr_tmp, i, min, max, 0.0, 1.0);
      
      if (img_list_te != img_list_tr)
	    {
	      img_list_temp = IMG_copy_list (img_list_te);
	      IMG_add_position_x_list (img_list_temp);
	      IMG_insert_in_list (img_list_temp, img_list_te_tmp, i);
	      IMG_free_list (img_list_temp);
	      IMG_norm_list (img_list_te_tmp, i, min, max, 0.0, 1.0);
	    }
      
      break;
    case IMG_position_y:

      img_list_temp=IMG_copy_list(img_list_tr);
      IMG_add_position_y_list(img_list_temp);
      IMG_insert_in_list(img_list_temp, img_list_tr_tmp,i);
      IMG_free_list(img_list_temp);
	  IMG_norm_list (img_list_tr_tmp, i, min, max, 0.0, 1.0);
      
      if (img_list_te != img_list_tr)
	    {
	      img_list_temp = IMG_copy_list (img_list_te);
	      IMG_add_position_y_list (img_list_temp);
	      IMG_insert_in_list (img_list_temp, img_list_te_tmp, i);
	      IMG_free_list (img_list_temp);
	      IMG_norm_list (img_list_te_tmp, i, min, max, 0.0, 1.0);
	    }
      
      break;

	case IMG_sobelh:
	case IMG_sobelv:
	case IMG_sobeld:
	case IMG_sobelhplus:
	case IMG_sobelvplus:
	case IMG_sobeldplus:
	case IMG_sobelhjaehne:
	case IMG_sobelvjaehne:
	case IMG_sobeldjaehne:
	case IMG_laplace:
	  min=-1.0;
	  max=+1.0;

	  img_list_temp = IMG_copy_list (img_list_tr);
	  IMG_filter_list (img_list_temp, dis_mode_list[(i * 3)]);
	  IMG_insert_in_list (img_list_temp, img_list_tr_tmp, i);
	  IMG_free_list (img_list_temp);

      if (img_list_te != img_list_tr)
	    {
	      img_list_temp = IMG_copy_list (img_list_te);
	      IMG_filter_list (img_list_temp, dis_mode_list[(i * 3)]);
	      IMG_insert_in_list (img_list_temp, img_list_te_tmp, i);
	      IMG_free_list (img_list_temp);
	      IMG_norm_list (img_list_te_tmp, i, min, max, 0.0, 1.0);
	    }

	  IMG_norm_list (img_list_tr_tmp, i, min, max, 0.0, 1.0);

	  break;
	}
      IMG_list_set_dis_mode (img_list_tr_tmp, i,
			     (int) dis_mode_list[(i * 3) + 1]);
      IMG_list_set_dis_scale (img_list_tr_tmp, i, dis_mode_list[(i * 3) + 2]);
      if (img_list_te != img_list_tr)
	{
	  IMG_list_set_dis_mode (img_list_te_tmp, i,
				 (int) dis_mode_list[(i * 3) + 1]);
	  IMG_list_set_dis_scale (img_list_te_tmp, i, dis_mode_list[(i * 3) + 2]);	  
	}
    }
  IMG_list_set_beamsize (img_list_tr_tmp, beamsize);
  IMG_list_set_dis_limit (img_list_tr_tmp, dis_limit);
  IMG_list_set_warprange (img_list_tr_tmp, warprange);
  if (img_list_te != img_list_tr)
    {
      IMG_list_set_beamsize (img_list_te_tmp, beamsize);
      IMG_list_set_dis_limit (img_list_te_tmp, dis_limit);
      IMG_list_set_warprange (img_list_te_tmp, warprange);
    }  


  if (img_list_te != img_list_tr)
    {
      IMG_free_list (img_list_te);
      img_list_te = img_list_te_tmp;
      IMG_print_list_info (img_list_te);
    }
  else
    {
      img_list_te = img_list_tr_tmp;
    }
  IMG_free_list (img_list_tr);
  img_list_tr = img_list_tr_tmp;
  IMG_print_list_info (img_list_tr);

//**********************************************

    switch (W2D_alg)
      {
      case W2D_W:
        W2D_function=W2DW_distance;
        break;
      case W2D_HM:
        W2D_function=W2DHM_distance;
        break;
      case W2D_HMD:
        W2D_function=W2DHMD_distance;
        break;
      case W2D_SA:
        W2D_function=W2DSA_distance;
        break;
      case W2D_IDM:
        W2D_function=W2DIDM_distance;
        break;
      case W2D_IDM3:
        W2D_function=W2DIDM3_distance;
        break;
      case W2D_HUNG:
        W2D_function=W2DHUNG_distance;        
        break;
      case W2D_TD:
        W2D_function=W2DTD_distance;        
        break;
      case W2D_NO:
        W2D_function=W2D_NOTHING;
        break;
      default:
        fprintf(stderr,"No algorithm selected\n"); 
        exit(-1);
        break;
      } 


//***************************************
//the next functions are used for learning  

    if(cluster_mode>=0)
      {
	fprintf(stdout,"#Cluster_mode: %d\n",cluster_mode);
	
	img_list_tr_tmp=IMG_create_mean_img_list(img_list_tr,W2D_function,cluster_mode);
	
	if (img_list_te != img_list_tr)
	  IMG_free_list (img_list_tr);
	img_list_tr = img_list_tr_tmp;
	leaving_one_out=0;
      }

    if(W2D_warpsum_mode>0)
      {
	//for(i=0;i<5;i++)
	//  {
	fprintf(stdout,"#W2D_warpsum calculation\n");
	W2D_calculate_warpsum(img_list_tr,img_list_te,W2D_function,0);
	//  }
      }
				  
	
    if(W2D_penvar_mode>0)
      {
	fprintf(stdout,"#W2D_penvar calculation\n");	    
	W2D_calculate_penvar(img_list_tr,img_list_te,W2D_function,0);	  
      }

    if(dis5x5_var>0)
      {   
	for(i=0;i<5;i++)
	  {
	    fprintf(stdout,"#Pixeldistance 5x5 variance calculation\n");
	    W2D_calculate_dis5x5_var(img_list_tr, img_list_te, W2D_function);
	  }
      }
//****************************************************


//needed for backtracking the ww_lists for visualation of the warped image

    if(IMG_get_max_dim_list(img_list_te)>=IMG_get_max_dim_list(img_list_tr))
      warping_info =
	malloc (sizeof (int) * IMG_get_max_dim_list(img_list_te) * 2);
    else
      warping_info =
	malloc (sizeof (int) * IMG_get_max_dim_list(img_list_tr) * 2);   
    if (warping_info == NULL)
      {
	fprintf (stderr, "No mem for warping_info\n");
	exit (-2);
      }
    for(i=0;i<IMG_get_max_dim_list(img_list_te) * 2;i++)
      warping_info[i]=0;

  //*****************************************************

  if(NN_counter==-1)
	NN_counter=img_list_tr->total_nr;
  NN_dis = malloc (sizeof (double) * (NN_counter + 1));
  NN_img = malloc (sizeof (int) * (NN_counter + 1));
  img_list_temp = malloc (sizeof(IMG_list));
  if(NN_dis == NULL || NN_img == NULL || img_list_temp == NULL)
    {
      fprintf (stderr, "not enough memory for NN_IMG_list\n");
      exit (-1);
    }
  img_list_temp->img = malloc (sizeof (IMG) * NN_counter);
  if( img_list_temp->img == NULL)
    {
      fprintf (stderr, "not enough memory for NN_IMG_list\n");
      exit (-1);
    }
  img_list_temp->total_nr = NN_counter;
  img_list_temp->name="temp_IMG_list";


  disdat_offset=0;
  for (i = test_img_start; i < img_list_te->total_nr; i++)
    {
      if(disdat_file_name!=NULL && disdat.row[disdat_offset].test_nr!=i)
	i=disdat.row[disdat_offset].test_nr;
      
      if (i >= test_img_counter + test_img_start && test_img_counter != -1)
	break;
 
      for (z = 0; z < img_list_temp->total_nr; z++)
	NN_dis[z] = DBL_MAX;     
      if(disdat_file_name==NULL)
	{
	  //*******search NN_counter best images with Nearest Neighbour******* 
	  NN_counter=0;
	  for (j = train_img_start; j < img_list_tr->total_nr; j++)
	    {
	      if (j >= train_img_counter + train_img_start
		  && train_img_counter != 0)
		break;
	      
	      if (leaving_one_out == 0 || strcmp(img_list_te->img[i].name, img_list_tr->img[j].name) != 0 )
		{
		  if(NN_counter<img_list_temp->total_nr)
		    NN_counter++;
		  NN_dis[NN_counter] =
		    W2DEUC_distance (img_list_te->img[i], img_list_tr->img[j], NULL);
		  if( NN_dis[NN_counter] == DBL_MAX)
		  {
			fprintf(stderr,"W2DEUC_distance=DBL_MAX this should not happen for NN sorting\n");
		 	exit(-1); 
		  }
		  NN_img[NN_counter] = j;
		  for (z = 0; z < NN_counter; z++)
		    {
		      if (NN_dis[NN_counter] < NN_dis[z])
			{
			  dis = NN_dis[NN_counter];
			  NN_dis[NN_counter] = NN_dis[z];
			  NN_dis[z] = dis;
			  dis = NN_img[NN_counter];
			  NN_img[NN_counter] = NN_img[z];
			  NN_img[z] = dis;
			}
		    }
		}
	    }
	}
      else
	{
	  //********set NN_list from disdat-file******************************
	  NN_counter=0;
	  for (j = 0; j < disdat.row[disdat_offset].test_counter; j++)
	    {
	      if(NN_counter<img_list_temp->total_nr)
		NN_counter++;	      
	      NN_dis[NN_counter] = disdat.row[disdat_offset+j].disfield[0];
	      NN_img[NN_counter] = disdat.row[disdat_offset+j].train_nr;
	      if(NN_img[NN_counter]>=img_list_tr->total_nr)
		{
		  fprintf(stderr,"Train-image-nr in disdat-file %s can not match/n",disdat_file_name);
		  exit(-1);
		}
	      for (z = 0; z < NN_counter; z++)
		{
		  if (NN_dis[NN_counter] < NN_dis[z])
		    {
		      dis = NN_dis[NN_counter];
		      NN_dis[NN_counter] = NN_dis[z];
		      NN_dis[z] = dis;
		      int_temp = NN_img[NN_counter];
		      NN_img[NN_counter] = NN_img[z];
		      NN_img[z] = int_temp;
		    }
		}	      
	    }
	  disdat_offset+=disdat.row[disdat_offset].test_counter;
	}
      if(dis_limit!=-1)
	{
	  img_list_temp->total_nr = 1; //check for diff. class
	  for (z = 0; z < NN_counter; z++)
	    {
	      img_list_temp->img[z]=img_list_tr->img[NN_img[z]];
	      if(img_list_tr->img[NN_img[0]].class != img_list_tr->img[NN_img[z]].class || NN_mode==0) //check for diff. class
		img_list_temp->total_nr = NN_counter; //check for diff. class
	    }
	  dis=W2DW_list_distance(img_list_te->img[i],img_list_temp,dis_limit, NULL);
	}
      else
	{
	  NN_best_class=-1; //check for diff. class
	  NN_best_dis=DBL_MAX; //check for diff. class

	  //***remove not wanted classes from NN_list***
	  int_temp=0;
	  for (z = 0; z < NN_counter; z++)
	    {
	      j = NN_img[z];
	      if((test_class==-1 || test_class==img_list_te->img[i].class)
		 && (train_class==-1 || train_class==img_list_tr->img[j].class))
		{
		  if(z!=int_temp)
		    {
		      NN_img[int_temp]=NN_img[z];
		      NN_dis[int_temp]=NN_dis[z];
		    }
		  int_temp++;
		}
	    }
	  NN_counter=int_temp;
	  //********************************************

	  for (z = 0; z < NN_counter; z++)
	    {
	      j = NN_img[z];
	      dis=DBL_MAX;
	      dis_wop=DBL_MAX;
	      if((NN_mode<1 || (j != -1 && NN_best_class != img_list_tr->img[j].class))//check for diff. class
		 && (test_class==-1 || test_class==img_list_te->img[i].class)
		 && (train_class==-1 || train_class==img_list_tr->img[j].class))
		{
		  A=&(img_list_te->img[i]);
		  B=&(img_list_tr->img[j]);

		  fprintf (stdout, "%s %s %d %d %g ", img_list_te->img[i].name, img_list_tr->img[j].name,
			   img_list_te->img[i].class, img_list_tr->img[j].class,
			   NN_dis[z]);

		  //calculate distance

		  dis = W2D_function (*A, *B, warping_info);
		  if(SYM_dis == 1)
		    {
		      dis +=
			W2D_function (*B, *A, NULL);
		      dis = dis/2;
		    }
		  
		  fprintf (stdout, "%g ",dis);

		  /*create images*/
	          #ifdef USE_GDlib	  
		  if(save_warped_image == 1)
		    {
		      snprintf (string_buffer, STRING_BUFFER_LENGTH, "class:%d-test:%s-train:%s-warped.png"
				,img_list_tr->img[j].class,img_list_te->img[i].name, img_list_tr->img[j].name);
		      IMG_GD_save_warped_img_png(*A, *B, warping_info, string_buffer);
		    }
		  #endif //USE_GDlib
 
		  //W2D_print_warp_data (warping_info,img_list_te->img[i].x_size,img_list_te->img[i].y_size);

		  dis_wop =
		    W2D_img_dis_without_penalty_from_warp (*A, *B, warping_info);		  

		  fprintf (stdout, "%g\n", dis_wop);

                  #ifdef USE_GDlib
		  if (write_gif == 1 || Xdraw == 1)
		    {
		      snprintf (string_buffer, STRING_BUFFER_LENGTH, "%s %s %f %f %f",img_list_te->img[i].name, img_list_tr->img[j].name, NN_dis[z],
			       dis, dis_wop);

		      IMG_GD_draw_warping_images (&img_gd, *A, *B, warping_info, string_buffer);
		      
		    }
		  #ifdef USE_X11lib
		  if (Xdraw == 1)
		    IMG_X_draw (&img_gd);
		  #endif //USE_X11lib
		  if (write_gif == 1)
		    {
		      snprintf (string_buffer, STRING_BUFFER_LENGTH, "test:%s-train:%s.png",img_list_te->img[i].name, img_list_tr->img[j].name);
		      IMG_GD_save (&img_gd, string_buffer);
		    }
		  #endif //USE_GDlib

		  if(wait_mode)
		    getchar();
		  
		  NN_img[z]=-1; //check for diff. class
		  
		  //check for diff. class**********************************************
		  if(NN_best_class!=img_list_tr->img[j].class && NN_mode>0)
		    z=0;
		  if(NN_mode==2)
		    {
		      if(dis_wop<NN_best_dis)
			{
			  NN_best_dis=dis;
			  NN_best_class=img_list_tr->img[j].class;
			}
		    }
		  else
		    {
		      if(dis<NN_best_dis)
			{
			  NN_best_dis=dis;
			  NN_best_class=img_list_tr->img[j].class;
			}		
		    }
		  //check for diff. class**********************************************
		}//if(j!=-1)
	    }
	}
      if(disdat_file_name!=NULL && disdat.counter<=disdat_offset)
	break; 
    }//(for i)
  fprintf (stdout, "#End_time: ");
  print_time_date (stdout);

  //****************************************************
  if(save_train_data_name!=0)
    IMG_save_list (img_list_tr,save_train_data_name);
  
  //*****************************************************

#ifdef USE_GDlib
  if (write_gif == 1 || Xdraw == 1)
    IMG_GD_destroy(&img_gd);	
#endif

  IMG_print_list_info (img_list_tr);

  if(disdat_file_name!=NULL)
    DIS_disdat_free(disdat);
  free (warping_info);
  if (img_list_te != img_list_tr)
    IMG_free_list (img_list_te);
  IMG_free_list (img_list_tr);
  free(img_list_temp->img);
  free(img_list_temp);
  free(NN_dis);
  free(NN_img);

  return (0);
}
